from __future__ import annotations
from typing import Optional, Generator, TYPE_CHECKING
import copy

if TYPE_CHECKING:
    from edsl.scenarios.Scenario import Scenario
    from edsl.scenarios.ScenarioList import ScenarioList


class DocumentChunker:
    def __init__(self, scenario: "Scenario"):
        self.scenario = scenario

    @staticmethod
    def _line_chunks(text, num_lines: int) -> Generator[str, None, None]:
        """Split a text into chunks of a given size.

        :param text: The text to split.
        :param num_lines: The number of lines in each chunk.

        Example:

        >>> list(DocumentChunker._line_chunks("This is a test.\\nThis is a test. This is a test.", 1))
        ['This is a test.', 'This is a test. This is a test.']
        """
        lines = text.split("\n")
        for i in range(0, len(lines), num_lines):
            chunk = "\n".join(lines[i : i + num_lines])
            yield chunk

    @staticmethod
    def _word_chunks(text, num_words: int) -> Generator[str, None, None]:
        """Split a text into chunks of a given size.

        :param text: The text to split.
        :param num_words: The number of words in each chunk.

        Example:

        >>> list(DocumentChunker._word_chunks("This is a test.", 2))
        ['This is', 'a test.']
        """
        words = text.split()
        for i in range(0, len(words), num_words):
            chunk = " ".join(words[i : i + num_words])
            yield chunk

    def chunk(
        self,
        field,
        num_words: Optional[int] = None,
        num_lines: Optional[int] = None,
        include_original=False,
        hash_original=False,
    ) -> ScenarioList:
        """Split a field into chunks of a given size.

        :param field: The field to split.
        :param num_words: The number of words in each chunk.
        :param num_lines: The number of lines in each chunk.
        :param include_original: Whether to include the original field in the new scenarios.
        :param hash_original: Whether to hash the original field in the new scenarios.

        If you specify `include_original=True`, the original field will be included in the new scenarios with an "_original" suffix.
        """
        from edsl.scenarios.ScenarioList import ScenarioList
        import hashlib

        if num_words is not None:
            chunks = list(self._word_chunks(self.scenario[field], num_words))

        if num_lines is not None:
            chunks = list(self._line_chunks(self.scenario[field], num_lines))

        if num_words is None and num_lines is None:
            raise ValueError("You must specify either num_words or num_lines.")

        if num_words is not None and num_lines is not None:
            raise ValueError(
                "You must specify either num_words or num_lines, but not both."
            )

        scenarios = []
        for i, chunk in enumerate(chunks):
            new_scenario = copy.deepcopy(self.scenario)
            new_scenario[field] = chunk
            new_scenario[field + "_chunk"] = i
            if include_original:
                if hash_original:
                    new_scenario[field + "_original"] = hashlib.md5(
                        self.scenario[field].encode()
                    ).hexdigest()
                else:
                    new_scenario[field + "_original"] = self.scenario[field]
            scenarios.append(new_scenario)
        return ScenarioList(scenarios)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Optional, Literal
from dataclasses import dataclass, asdict

# from edsl.data_transfer_models import VisibilityType
from edsl.data.Cache import Cache
from edsl.jobs.buckets.BucketCollection import BucketCollection
from edsl.language_models.key_management.KeyLookup import KeyLookup
from edsl.jobs.runners.JobsRunnerStatus import JobsRunnerStatus

VisibilityType = Literal["private", "public", "unlisted"]
from edsl.Base import Base


@dataclass
class RunEnvironment:
    cache: Optional[Cache] = None
    bucket_collection: Optional[BucketCollection] = None
    key_lookup: Optional[KeyLookup] = None
    jobs_runner_status: Optional["JobsRunnerStatus"] = None


@dataclass
class RunParameters(Base):
    n: int = 1
    progress_bar: bool = False
    stop_on_exception: bool = False
    check_api_keys: bool = False
    verbose: bool = True
    print_exceptions: bool = True
    remote_cache_description: Optional[str] = None
    remote_inference_description: Optional[str] = None
    remote_inference_results_visibility: Optional[VisibilityType] = "unlisted"
    skip_retry: bool = False
    raise_validation_errors: bool = False
    disable_remote_cache: bool = False
    disable_remote_inference: bool = False
    job_uuid: Optional[str] = None

    def to_dict(self, add_edsl_version=False) -> dict:
        d = asdict(self)
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "RunConfig"
        return d

    @classmethod
    def from_dict(cls, data: dict) -> "RunConfig":
        return cls(**data)

    def code(self):
        return f"RunConfig(**{self.to_dict()})"

    @classmethod
    def example(cls) -> "RunConfig":
        return cls()


@dataclass
class RunConfig:
    environment: RunEnvironment
    parameters: RunParameters

    def add_environment(self, environment: RunEnvironment):
        self.environment = environment

    def add_bucket_collection(self, bucket_collection: BucketCollection):
        self.environment.bucket_collection = bucket_collection

    def add_cache(self, cache: Cache):
        self.environment.cache = cache

    def add_key_lookup(self, key_lookup: KeyLookup):
        self.environment.key_lookup = key_lookup


"""This module contains the Answers class, which is a helper class to hold the answers to a survey."""

from collections import UserDict
from edsl.data_transfer_models import EDSLResultObjectInput


class Answers(UserDict):
    """Helper class to hold the answers to a survey."""

    def add_answer(
        self, response: EDSLResultObjectInput, question: "QuestionBase"
    ) -> None:
        """Add a response to the answers dictionary."""
        answer = response.answer
        comment = response.comment
        generated_tokens = response.generated_tokens
        # record the answer
        if generated_tokens:
            self[question.question_name + "_generated_tokens"] = generated_tokens
        self[question.question_name] = answer
        if comment:
            self[question.question_name + "_comment"] = comment

    def replace_missing_answers_with_none(self, survey: "Survey") -> None:
        """Replace missing answers with None. Answers can be missing if the agent skips a question."""
        for question_name in survey.question_names:
            if question_name not in self:
                self[question_name] = None

    def to_dict(self):
        """Return a dictionary of the answers."""
        return self.data

    @classmethod
    def from_dict(cls, d):
        """Return an Answers object from a dictionary."""
        return cls(d)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import random


def default_turn_taking_generator(agent_list, speakers_so_far, **kwargs):
    """Returns the next speaker in the list of agents, in order. If no speakers have spoken yet, returns the first agent in the list."""
    if len(speakers_so_far) == 0:
        return agent_list[0]
    most_recent_speaker = speakers_so_far[-1]
    index = agent_list.index(most_recent_speaker)
    return agent_list[(index + 1) % len(agent_list)]


def turn_taking_generator_with_focal_speaker(
    agent_list, speakers_so_far, focal_speaker_index, **kwargs
):
    """Returns the focal speaker first, then the next speaker in the list of agents sequentially, going back to the focal speaker after the last agent has spoken. If no speakers have spoken yet, returns the focal speaker.
    This would be appropriate for say, a an auction where the auctioneer always speaks first, and then the next person in line speaks, and then the auctioneer speaks again.
    """
    if len(speakers_so_far) == 0:
        return agent_list[focal_speaker_index]

    most_recent_speaker = speakers_so_far[-1]
    if most_recent_speaker == agent_list[focal_speaker_index]:
        non_focal_agents = [
            a for a in agent_list if a != agent_list[focal_speaker_index]
        ]
        non_focal_speakers = [
            a for a in speakers_so_far if a != agent_list[focal_speaker_index]
        ]
        return default_turn_taking_generator(
            agent_list=non_focal_agents, speakers_so_far=non_focal_speakers
        )
    else:
        return agent_list[focal_speaker_index]


def random_turn_taking_generator(agent_list, speakers_so_far, **kwargs):
    """Returns a random speaker from the list of agents, but ensuring no agent speaks twice in a row.
    If no speakers have spoken yet, returns a random agent."""
    if len(speakers_so_far) == 0:
        return random.choice(agent_list)
    else:
        most_recent_speaker = speakers_so_far[-1]
    return random.choice([a for a in agent_list if a != most_recent_speaker])


def random_inclusive_generator(agent_list, speakers_so_far, **kwargs):
    """Returns a random speaker from the list of agents, but ensuring no agent speaks twice in a row and that
    every agent has spoken before the same agent speaks again.
    """
    if len(speakers_so_far) > 0:
        most_recent_speaker = speakers_so_far[-1]
    else:
        most_recent_speaker = None

    lookback_length = len(speakers_so_far) % len(agent_list)

    if lookback_length == 0:
        eligible_agents = agent_list[:]
    else:
        eligible_agents = [
            a for a in agent_list if a not in speakers_so_far[-lookback_length:]
        ]

    # don't have the same speaker twice in a row
    if most_recent_speaker in eligible_agents:
        eligible_agents.pop(eligible_agents.index(most_recent_speaker))

    return random.choice(eligible_agents)


def speaker_closure(agent_list, generator_function, focal_speaker_index=None):
    speakers_so_far = []
    focal_speaker_index = focal_speaker_index

    def next_speaker_generator():
        speaker = generator_function(
            agent_list=agent_list,
            speakers_so_far=speakers_so_far,
            focal_speaker_index=focal_speaker_index,
        )
        speakers_so_far.append(speaker)
        return speaker

    return next_speaker_generator


# speaker_gen = speaker_closure(agent_list = ['Alice', 'Bob', 'Charlie', "Mr.X"],
#                               generator_function = turn_taking_generator_with_focal_speaker,
#                               focal_speaker_index = 3)

# for _ in range(30):
#     print(speaker_gen())



"""A mixin class for exporting surveys to different formats."""

from typing import Union, Optional

import subprocess
import platform
import os
import tempfile


def open_docx(file_path):
    """
    Open a docx file using the default application in a cross-platform manner.

    :param file_path: str, path to the docx file
    """
    file_path = os.path.abspath(file_path)

    if platform.system() == "Darwin":  # macOS
        subprocess.call(("open", file_path))
    elif platform.system() == "Windows":  # Windows
        os.startfile(file_path)
    else:  # linux variants
        subprocess.call(("xdg-open", file_path))


class SurveyExportMixin:
    """A mixin class for exporting surveys to different formats."""

    def css(self):
        from edsl.surveys.SurveyCSS import SurveyCSS

        return SurveyCSS.default_style().generate_css()

    def get_description(self) -> str:
        """Return the description of the survey."""
        from edsl import QuestionFreeText

        question_texts = "\n".join([q.question_text for q in self._questions])
        q = QuestionFreeText(
            question_name="description",
            question_text=f"""A survey was conducted with the following questions: 
                             {question_texts}
                             Please write a description of the survey.
                             """,
        )
        return q.run().select("description").first()

    def docx(
        self,
        return_document_object: bool = False,
        filename: Optional[str] = None,
        open_file: bool = False,
    ) -> Union["Document", None]:
        """Generate a docx document for the survey."""
        from docx import Document

        doc = Document()
        doc.add_heading("EDSL Survey")
        doc.add_paragraph(f"\n")
        for index, question in enumerate(self._questions):
            h = doc.add_paragraph()  # Add question as a paragraph
            h.add_run(f"Question {index + 1} ({question.question_name})").bold = True
            h.add_run(f"; {question.question_type}").italic = True
            p = doc.add_paragraph()
            p.add_run(question.question_text)
            if question.question_type == "linear_scale":
                for key, value in getattr(question, "option_labels", {}).items():
                    doc.add_paragraph(str(key) + ": " + str(value), style="ListBullet")
            else:
                if hasattr(question, "question_options"):
                    for option in getattr(question, "question_options", []):
                        doc.add_paragraph(str(option), style="ListBullet")

        if return_document_object and filename is None:
            return doc

        if filename is None:
            with tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".docx", dir=os.getcwd()
            ) as f:
                filename = f.name

        doc.save(filename)
        print("The survey has been saved to", filename)
        if open_file:
            open_docx(filename)
        return

    def show(self):
        self.to_scenario_list(questions_only=False, rename=True).print(format="rich")

    def to_scenario_list(
        self, questions_only: bool = True, rename=False
    ) -> "ScenarioList":
        from edsl import ScenarioList, Scenario

        # from edsl.questions import QuestionBase

        if questions_only:
            to_iterate_over = self._questions
        else:
            to_iterate_over = self.recombined_questions_and_instructions()

        if rename:
            renaming_dict = {
                "name": "identifier",
                "question_name": "identifier",
                "question_text": "text",
            }
        else:
            renaming_dict = {}

        all_keys = set([])
        scenarios = ScenarioList()
        for item in to_iterate_over:
            d = item.to_dict()
            if item.__class__.__name__ == "Instruction":
                d["question_type"] = "NA / instruction"
            for key in renaming_dict:
                if key in d:
                    d[renaming_dict[key]] = d.pop(key)
            all_keys.update(d.keys())
            scenarios.append(Scenario(d))

        for scenario in scenarios:
            for key in all_keys:
                if key not in scenario:
                    scenario[key] = None

        return scenarios

    def code(self, filename: str = None, survey_var_name: str = "survey") -> list[str]:
        """Create the Python code representation of a survey.

        :param filename: The name of the file to save the code to.
        :param survey_var_name: The name of the survey variable.

        >>> from edsl.surveys import Survey
        >>> survey = Survey.example()
        >>> print(survey.code())
        from edsl.surveys.Survey import Survey
        ...
        ...
        survey = Survey(questions=[q0, q1, q2])
        ...
        """
        import black

        header_lines = ["from edsl.surveys.Survey import Survey"]
        header_lines.append("from edsl import Question")
        lines = ["\n".join(header_lines)]
        for question in self._questions:
            question.question_text = question["question_text"].replace("\n", " ")
            # remove dublicate spaces
            question.question_text = " ".join(question.question_text.split())
            lines.append(f"{question.question_name} = " + repr(question))
        lines.append(
            f"{survey_var_name} = Survey(questions = [{', '.join(self.question_names)}])"
        )
        # return lines
        code_string = "\n".join(lines)
        formatted_code = black.format_str(code_string, mode=black.FileMode())

        if filename:
            print("The code has been saved to", filename)
            print("The survey itself is saved to 'survey' object")
            with open(filename, "w") as file:
                file.write(formatted_code)
            return

        return formatted_code

    def html(
        self,
        scenario: Optional[dict] = None,
        filename: Optional[str] = None,
        return_link=False,
        css: Optional[str] = None,
        cta: Optional[str] = "Open HTML file",
        include_question_name=False,
    ):
        from IPython.display import display, HTML
        import tempfile
        import os
        from edsl.utilities.utilities import is_notebook

        if scenario is None:
            scenario = {}

        if css is None:
            css = self.css()

        if filename is None:
            current_directory = os.getcwd()
            filename = tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".html", dir=current_directory
            ).name

        html_header = f"""<html>
        <head><title></title>
        <style>
        { css }
        </style>
        </head>
        <body>
        <div class="survey_container">
        """

        html_footer = """
        </div>
        </body>
        </html>"""

        output = html_header

        with open(filename, "w") as f:
            f.write(html_header)
            for question in self._questions:
                f.write(
                    question.html(
                        scenario=scenario, include_question_name=include_question_name
                    )
                )
                output += question.html(
                    scenario=scenario, include_question_name=include_question_name
                )
            f.write(html_footer)
            output += html_footer

        if is_notebook():
            html_url = f"/files/{filename}"
            html_link = f'<a href="{html_url}" target="_blank">{cta}</a>'
            display(HTML(html_link))

            import html

            escaped_output = html.escape(output)
            iframe = f""""
            <iframe srcdoc="{ escaped_output }" style="width: 800px; height: 600px;"></iframe>
            """
            display(HTML(iframe))

        else:
            print(f"Survey saved to {filename}")
            import webbrowser
            import os

            webbrowser.open(f"file://{os.path.abspath(filename)}")
            # webbrowser.open(filename)

        if return_link:
            return filename


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import base64
import io
import tempfile
import mimetypes
import os
from typing import Dict, Any, IO, Optional

from edsl.scenarios.Scenario import Scenario
from edsl.utilities.remove_edsl_version import remove_edsl_version

from edsl.scenarios.file_methods import FileMethods


class FileStore(Scenario):
    __documentation__ = "https://docs.expectedparrot.com/en/latest/filestore.html"

    def __init__(
        self,
        path: Optional[str] = None,
        mime_type: Optional[str] = None,
        binary: Optional[bool] = None,
        suffix: Optional[str] = None,
        base64_string: Optional[str] = None,
        external_locations: Optional[Dict[str, str]] = None,
        extracted_text: Optional[str] = None,
        **kwargs,
    ):
        if path is None and "filename" in kwargs:
            path = kwargs["filename"]

        self._path = path  # Store the original path privately
        self._temp_path = None  # Track any generated temporary file

        self.suffix = suffix or path.split(".")[-1]
        self.binary = binary or False
        self.mime_type = (
            mime_type or mimetypes.guess_type(path)[0] or "application/octet-stream"
        )
        self.base64_string = base64_string or self.encode_file_to_base64_string(path)
        self.external_locations = external_locations or {}

        self.extracted_text = (
            self.extract_text() if extracted_text is None else extracted_text
        )

        super().__init__(
            {
                "path": path,
                "base64_string": self.base64_string,
                "binary": self.binary,
                "suffix": self.suffix,
                "mime_type": self.mime_type,
                "external_locations": self.external_locations,
                "extracted_text": self.extracted_text,
            }
        )

    @property
    def path(self) -> str:
        """
        Property that returns a valid path to the file content.
        If the original path doesn't exist, generates a temporary file from the base64 content.
        """
        # Check if original path exists and is accessible
        if self._path and os.path.isfile(self._path):
            return self._path

        # If we already have a valid temporary file, use it
        if self._temp_path and os.path.isfile(self._temp_path):
            return self._temp_path

        # Generate a new temporary file from base64 content
        self._temp_path = self.to_tempfile(self.suffix)
        return self._temp_path

    def __str__(self):
        return "FileStore: self.path"

    @classmethod
    def example(cls, example_type="txt"):
        file_methods_class = FileMethods.get_handler(example_type)
        if file_methods_class:
            return cls(file_methods_class().example())
        else:
            print(f"Example for {example_type} is not supported.")

    @property
    def size(self) -> int:
        if self.base64_string != None:
            return (len(self.base64_string) / 4.0) * 3  # from base64 to char size
        return os.path.getsize(self.path)

    def upload_google(self, refresh: bool = False) -> None:
        import google.generativeai as genai

        genai.configure(api_key=os.getenv("GOOGLE_API_KEY"))
        google_info = genai.upload_file(self.path, mime_type=self.mime_type)
        self.external_locations["google"] = google_info.to_dict()

    @classmethod
    @remove_edsl_version
    def from_dict(cls, d):
        # return cls(d["filename"], d["binary"], d["suffix"], d["base64_string"])
        return cls(**d)

    def __repr__(self):
        import reprlib

        r = reprlib.Repr()
        r.maxstring = 20  # Limit strings to 20 chars
        r.maxother = 30  # Limit other types to 30 chars

        params = ", ".join(f"{key}={r.repr(value)}" for key, value in self.data.items())
        return f"{self.__class__.__name__}({params})"

    def _repr_html_(self):
        parent_html = super()._repr_html_()
        from edsl.scenarios.ConstructDownloadLink import ConstructDownloadLink

        link = ConstructDownloadLink(self).html_create_link(self.path, style=None)
        return f"{parent_html}<br>{link}"

    def encode_file_to_base64_string(self, file_path: str):
        try:
            # Attempt to open the file in text mode
            with open(file_path, "r") as text_file:
                # Read the text data
                text_data = text_file.read()
                # Encode the text data to a base64 string
                base64_encoded_data = base64.b64encode(text_data.encode("utf-8"))
        except UnicodeDecodeError:
            # If reading as text fails, open the file in binary mode
            with open(file_path, "rb") as binary_file:
                # Read the binary data
                binary_data = binary_file.read()
                # Encode the binary data to a base64 string
                base64_encoded_data = base64.b64encode(binary_data)
                self.binary = True
        # Convert the base64 bytes to a string
        base64_string = base64_encoded_data.decode("utf-8")

        return base64_string

    def open(self) -> "IO":
        if self.binary:
            return self.base64_to_file(self.base64_string, is_binary=True)
        else:
            return self.base64_to_text_file(self.base64_string)

    def write(self, filename: Optional[str] = None) -> str:
        """
        Write the file content to disk, either to a specified filename or a temporary file.

        Args:
            filename (Optional[str]): The destination filename. If None, creates a temporary file.

        Returns:
            str: The path to the written file.
        """
        # Determine the mode based on binary flag
        mode = "wb" if self.binary else "w"

        # If no filename provided, create a temporary file
        if filename is None:
            from tempfile import NamedTemporaryFile

            with NamedTemporaryFile(delete=False, suffix="." + self.suffix) as f:
                filename = f.name

        # Write the content using the appropriate mode
        try:
            with open(filename, mode) as f:
                content = self.open().read()
                # For text mode, ensure we're writing a string
                if not self.binary and isinstance(content, bytes):
                    content = content.decode("utf-8")
                f.write(content)
                print(f"File written to {filename}")
        except Exception as e:
            print(f"Error writing file: {e}")
            raise

        # return filename

    @staticmethod
    def base64_to_text_file(base64_string) -> "IO":
        # Decode the base64 string to bytes
        text_data_bytes = base64.b64decode(base64_string)

        # Convert bytes to string
        text_data = text_data_bytes.decode("utf-8")

        # Create a StringIO object from the text data
        text_file = io.StringIO(text_data)

        return text_file

    @staticmethod
    def base64_to_file(base64_string, is_binary=True):
        # Decode the base64 string to bytes
        file_data = base64.b64decode(base64_string)

        if is_binary:
            # Create a BytesIO object for binary data
            return io.BytesIO(file_data)
        else:
            # Convert bytes to string for text data
            text_data = file_data.decode("utf-8")
            # Create a StringIO object for text data
            return io.StringIO(text_data)

    @property
    def text(self):
        if self.binary:
            import warnings

            warnings.warn("This is a binary file.")
        else:
            return self.base64_to_text_file(self.base64_string).read()

    def to_tempfile(self, suffix=None):
        if suffix is None:
            suffix = self.suffix
        if self.binary:
            file_like_object = self.base64_to_file(
                self["base64_string"], is_binary=True
            )
        else:
            file_like_object = self.base64_to_text_file(self.base64_string)

        # Create a named temporary file
        mode = "wb" if self.binary else "w"
        temp_file = tempfile.NamedTemporaryFile(
            delete=False, suffix="." + suffix, mode=mode
        )

        if self.binary:
            temp_file.write(file_like_object.read())
        else:
            temp_file.write(file_like_object.read())

        temp_file.close()

        return temp_file.name

    def view(self) -> None:
        handler = FileMethods.get_handler(self.suffix)
        if handler:
            handler(self.path).view()
        else:
            print(f"Viewing of {self.suffix} files is not supported.")

    def extract_text(self) -> str:
        handler = FileMethods.get_handler(self.suffix)
        if handler and hasattr(handler, "extract_text"):
            return handler(self.path).extract_text()

        if not self.binary:
            return self.text

        return None
        # raise TypeError("No text method found for this file type.")

    def push(
        self, description: Optional[str] = None, visibility: str = "unlisted"
    ) -> dict:
        """
        Push the object to Coop.
        :param description: The description of the object to push.
        :param visibility: The visibility of the object to push.
        """
        scenario_version = Scenario.from_dict(self.to_dict())
        if description is None:
            description = "File: " + self.path
        info = scenario_version.push(description=description, visibility=visibility)
        return info

    @classmethod
    def pull(cls, uuid: str, expected_parrot_url: Optional[str] = None) -> "FileStore":
        """
        :param uuid: The UUID of the object to pull.
        :param expected_parrot_url: The URL of the Parrot server to use.
        :return: The object pulled from the Parrot server.
        """
        scenario_version = Scenario.pull(uuid, expected_parrot_url=expected_parrot_url)
        return cls.from_dict(scenario_version.to_dict())

    @classmethod
    def from_url(
        cls,
        url: str,
        download_path: Optional[str] = None,
        mime_type: Optional[str] = None,
    ) -> "FileStore":
        """
        :param url: The URL of the file to download.
        :param download_path: The path to save the downloaded file.
        :param mime_type: The MIME type of the file. If None, it will be guessed from the file extension.
        """
        import requests
        from urllib.parse import urlparse

        response = requests.get(url, stream=True)
        response.raise_for_status()  # Raises an HTTPError for bad responses

        # Get the filename from the URL if download_path is not provided
        if download_path is None:
            filename = os.path.basename(urlparse(url).path)
            if not filename:
                filename = "downloaded_file"
            # download_path = filename
            download_path = os.path.join(os.getcwd(), filename)

        # Ensure the directory exists
        os.makedirs(os.path.dirname(download_path), exist_ok=True)

        # Write the file
        with open(download_path, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        # Create and return a new File instance
        return cls(download_path, mime_type=mime_type)

    def create_link(self, custom_filename=None, style=None):
        from edsl.scenarios.ConstructDownloadLink import ConstructDownloadLink

        return ConstructDownloadLink(self).create_link(custom_filename, style)


class CSVFileStore(FileStore):
    @classmethod
    def example(cls):
        from edsl.results.Results import Results

        r = Results.example()
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".csv", delete=False) as f:
            r.to_csv(filename=f.name)

        return cls(f.name)

    def view(self):
        import pandas as pd

        return pd.read_csv(self.to_tempfile())


class PDFFileStore(FileStore):
    def view(self):
        pdf_path = self.to_tempfile()
        print(f"PDF path: {pdf_path}")  # Print the path to ensure it exists
        import os
        import subprocess

        if os.path.exists(pdf_path):
            try:
                if os.name == "posix":
                    # for cool kids
                    subprocess.run(["open", pdf_path], check=True)  # macOS
                elif os.name == "nt":
                    os.startfile(pdf_path)  # Windows
                else:
                    subprocess.run(["xdg-open", pdf_path], check=True)  # Linux
            except Exception as e:
                print(f"Error opening PDF: {e}")
        else:
            print("PDF file was not created successfully.")

    @classmethod
    def example(cls):
        import textwrap

        pdf_string = textwrap.dedent(
            """\
        %PDF-1.4
        1 0 obj
        << /Type /Catalog /Pages 2 0 R >>
        endobj
        2 0 obj
        << /Type /Pages /Kids [3 0 R] /Count 1 >>
        endobj
        3 0 obj
        << /Type /Page /Parent 2 0 R /MediaBox [0 0 612 792] /Contents 4 0 R >>
        endobj
        4 0 obj
        << /Length 44 >>
        stream
        BT
        /F1 24 Tf
        100 700 Td
        (Hello, World!) Tj
        ET
        endstream
        endobj
        5 0 obj
        << /Type /Font /Subtype /Type1 /BaseFont /Helvetica >>
        endobj
        6 0 obj
        << /ProcSet [/PDF /Text] /Font << /F1 5 0 R >> >>
        endobj
        xref
        0 7
        0000000000 65535 f 
        0000000010 00000 n 
        0000000053 00000 n 
        0000000100 00000 n 
        0000000173 00000 n 
        0000000232 00000 n 
        0000000272 00000 n 
        trailer
        << /Size 7 /Root 1 0 R >>
        startxref
        318
        %%EOF"""
        )
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".pdf", delete=False) as f:
            f.write(pdf_string.encode())

        return cls(f.name)


class PNGFileStore(FileStore):
    @classmethod
    def example(cls):
        import textwrap

        png_string = textwrap.dedent(
            """\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x01\x00\x00\x00\x01\x00\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\x0cIDAT\x08\xd7c\x00\x01"""
        )
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as f:
            f.write(png_string.encode())

        return cls(f.name)

    def view(self):
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg

        img = mpimg.imread(self.to_tempfile())
        plt.imshow(img)
        plt.show()


class SQLiteFileStore(FileStore):
    @classmethod
    def example(cls):
        import sqlite3
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".sqlite", delete=False) as f:
            conn = sqlite3.connect(f.name)
            c = conn.cursor()
            c.execute("""CREATE TABLE stocks (date text)""")
            conn.commit()

            return cls(f.name)

    def view(self):
        import subprocess
        import os

        sqlite_path = self.to_tempfile()
        os.system(f"sqlite3 {sqlite_path}")


class HTMLFileStore(FileStore):
    @classmethod
    def example(cls):
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as f:
            f.write("<html><body><h1>Test</h1></body></html>".encode())

        return cls(f.name)

    def view(self):
        import webbrowser

        html_path = self.to_tempfile()
        webbrowser.open("file://" + html_path)


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    # fs = FileStore.example("pdf")
    # fs.view()

    formats = FileMethods.supported_file_types()
    for file_type in formats:
        print("Now testinging", file_type)
        fs = FileStore.example(file_type)
        fs.view()
        input("Press Enter to continue...")

    # pdf_example.view()
    # FileStore(pdf_example).view()

    # pdf_methods = methods.get("pdf")
    # file = pdf_methods().example()
    # pdf_methods(file).view()

    # print(FileMethods._handlers)



"""This module contains the LanguageModel class, which is an abstract base class for all language models.

Terminology:

raw_response: The JSON response from the model. This has all the model meta-data about the call. 

edsl_augmented_response: The JSON response from model, but augmented with EDSL-specific information, 
such as the cache key, token usage, etc. 

generated_tokens: The actual tokens generated by the model. This is the output that is used by the user.
edsl_answer_dict: The parsed JSON response from the model either {'answer': ...} or {'answer': ..., 'comment': ...}

"""

from __future__ import annotations
import warnings
from functools import wraps
import asyncio
import json
import os
from typing import (
    Coroutine,
    Any,
    Type,
    Union,
    List,
    get_type_hints,
    TypedDict,
    Optional,
    TYPE_CHECKING,
)
from abc import ABC, abstractmethod

from edsl.data_transfer_models import (
    ModelResponse,
    ModelInputs,
    EDSLOutput,
    AgentResponseDict,
)

if TYPE_CHECKING:
    from edsl.data.Cache import Cache
    from edsl.scenarios.FileStore import FileStore
    from edsl.questions.QuestionBase import QuestionBase
    from edsl.language_models.key_management.KeyLookup import KeyLookup

from edsl.enums import InferenceServiceType

from edsl.utilities.decorators import (
    sync_wrapper,
    jupyter_nb_handler,
)
from edsl.utilities.remove_edsl_version import remove_edsl_version

from edsl.Base import PersistenceMixin, RepresentationMixin
from edsl.language_models.RegisterLanguageModelsMeta import RegisterLanguageModelsMeta

from edsl.language_models.key_management.KeyLookupCollection import (
    KeyLookupCollection,
)

from edsl.language_models.RawResponseHandler import RawResponseHandler


def handle_key_error(func):
    """Handle KeyError exceptions."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
            assert True == False
        except KeyError as e:
            return f"""KeyError occurred: {e}. This is most likely because the model you are using 
            returned a JSON object we were not expecting."""

    return wrapper


class classproperty:
    def __init__(self, method):
        self.method = method

    def __get__(self, instance, cls):
        return self.method(cls)


from edsl.Base import HashingMixin


class LanguageModel(
    PersistenceMixin,
    RepresentationMixin,
    HashingMixin,
    ABC,
    metaclass=RegisterLanguageModelsMeta,
):
    """ABC for Language Models."""

    _model_ = None
    key_sequence = (
        None  # This should be something like ["choices", 0, "message", "content"]
    )

    DEFAULT_RPM = 100
    DEFAULT_TPM = 1000

    @classproperty
    def response_handler(cls):
        key_sequence = cls.key_sequence
        usage_sequence = cls.usage_sequence if hasattr(cls, "usage_sequence") else None
        return RawResponseHandler(key_sequence, usage_sequence)

    def __init__(
        self,
        tpm: Optional[float] = None,
        rpm: Optional[float] = None,
        omit_system_prompt_if_empty_string: bool = True,
        key_lookup: Optional["KeyLookup"] = None,
        **kwargs,
    ):
        """Initialize the LanguageModel."""
        self.model = getattr(self, "_model_", None)
        default_parameters = getattr(self, "_parameters_", None)
        parameters = self._overide_default_parameters(kwargs, default_parameters)
        self.parameters = parameters
        self.remote = False
        self.omit_system_prompt_if_empty = omit_system_prompt_if_empty_string

        self.key_lookup = self._set_key_lookup(key_lookup)
        self.model_info = self.key_lookup.get(self._inference_service_)

        if rpm is not None:
            self._rpm = rpm

        if tpm is not None:
            self._tpm = tpm

        for key, value in parameters.items():
            setattr(self, key, value)

        for key, value in kwargs.items():
            if key not in parameters:
                setattr(self, key, value)

        if kwargs.get("skip_api_key_check", False):
            # Skip the API key check. Sometimes this is useful for testing.
            self._api_token = None

    def _set_key_lookup(self, key_lookup: "KeyLookup") -> "KeyLookup":
        """Set the key lookup."""
        if key_lookup is not None:
            return key_lookup
        else:
            klc = KeyLookupCollection()
            klc.add_key_lookup(fetch_order=("config", "env"))
            return klc.get(("config", "env"))

    def set_key_lookup(self, key_lookup: "KeyLookup") -> None:
        """Set the key lookup, later"""
        if hasattr(self, "_api_token"):
            del self._api_token
        self.key_lookup = key_lookup

    def ask_question(self, question: "QuestionBase") -> str:
        """Ask a question and return the response.

        :param question: The question to ask.
        """
        user_prompt = question.get_instructions().render(question.data).text
        system_prompt = "You are a helpful agent pretending to be a human."
        return self.execute_model_call(user_prompt, system_prompt)

    @property
    def rpm(self):
        if not hasattr(self, "_rpm"):
            if self.model_info is None:
                self._rpm = self.DEFAULT_RPM
            else:
                self._rpm = self.model_info.rpm
        return self._rpm

    @property
    def tpm(self):
        if not hasattr(self, "_tpm"):
            if self.model_info is None:
                self._tpm = self.DEFAULT_TPM
            else:
                self._tpm = self.model_info.tpm
        return self._tpm

    # in case we want to override the default values
    @tpm.setter
    def tpm(self, value):
        self._tpm = value

    @rpm.setter
    def rpm(self, value):
        self._rpm = value

    @property
    def api_token(self) -> str:
        if not hasattr(self, "_api_token"):
            info = self.key_lookup.get(self._inference_service_, None)
            if info is None:
                raise ValueError(
                    f"No key found for service '{self._inference_service_}'"
                )
            self._api_token = info.api_token
        return self._api_token

    def __getitem__(self, key):
        return getattr(self, key)

    def hello(self, verbose=False):
        """Runs a simple test to check if the model is working."""
        token = self.api_token
        masked = token[: min(8, len(token))] + "..."
        if verbose:
            print(f"Current key is {masked}")
        return self.execute_model_call(
            user_prompt="Hello, model!", system_prompt="You are a helpful agent."
        )

    def has_valid_api_key(self) -> bool:
        """Check if the model has a valid API key.

        >>> LanguageModel.example().has_valid_api_key() : # doctest: +SKIP
        True

        This method is used to check if the model has a valid API key.
        """
        from edsl.enums import service_to_api_keyname

        if self._model_ == "test":
            return True

        key_name = service_to_api_keyname.get(self._inference_service_, "NOT FOUND")
        key_value = os.getenv(key_name)
        return key_value is not None

    def __hash__(self) -> str:
        """Allow the model to be used as a key in a dictionary.

        >>> m = LanguageModel.example()
        >>> hash(m)
        1811901442659237949
        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False))

    def __eq__(self, other) -> bool:
        """Check is two models are the same.

        >>> m1 = LanguageModel.example()
        >>> m2 = LanguageModel.example()
        >>> m1 == m2
        True

        """
        return self.model == other.model and self.parameters == other.parameters

    @staticmethod
    def _overide_default_parameters(passed_parameter_dict, default_parameter_dict):
        """Return a dictionary of parameters, with passed parameters taking precedence over defaults.

        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9})
        {'temperature': 0.5}
        >>> LanguageModel._overide_default_parameters(passed_parameter_dict={"temperature": 0.5}, default_parameter_dict={"temperature":0.9, "max_tokens": 1000})
        {'temperature': 0.5, 'max_tokens': 1000}
        """
        # this is the case when data is loaded from a dict after serialization
        if "parameters" in passed_parameter_dict:
            passed_parameter_dict = passed_parameter_dict["parameters"]
        return {
            parameter_name: passed_parameter_dict.get(parameter_name, default_value)
            for parameter_name, default_value in default_parameter_dict.items()
        }

    def __call__(self, user_prompt: str, system_prompt: str):
        return self.execute_model_call(user_prompt, system_prompt)

    @abstractmethod
    async def async_execute_model_call(user_prompt: str, system_prompt: str):
        """Execute the model call and returns a coroutine."""
        pass

    async def remote_async_execute_model_call(
        self, user_prompt: str, system_prompt: str
    ):
        """Execute the model call and returns the result as a coroutine, using Coop."""
        from edsl.coop import Coop

        client = Coop()
        response_data = await client.remote_async_execute_model_call(
            self.to_dict(), user_prompt, system_prompt
        )
        return response_data

    @jupyter_nb_handler
    def execute_model_call(self, *args, **kwargs) -> Coroutine:
        """Execute the model call and returns the result as a coroutine."""

        async def main():
            results = await asyncio.gather(
                self.async_execute_model_call(*args, **kwargs)
            )
            return results[0]  # Since there's only one task, return its result

        return main()

    @classmethod
    def get_generated_token_string(cls, raw_response: dict[str, Any]) -> str:
        """Return the generated token string from the raw response.

        >>> m = LanguageModel.example(test_model = True)
        >>> raw_response = m.execute_model_call("Hello, model!", "You are a helpful agent.")
        >>> m.get_generated_token_string(raw_response)
        'Hello world'

        """
        return cls.response_handler.get_generated_token_string(raw_response)

    @classmethod
    def get_usage_dict(cls, raw_response: dict[str, Any]) -> dict[str, Any]:
        """Return the usage dictionary from the raw response."""
        return cls.response_handler.get_usage_dict(raw_response)

    @classmethod
    def parse_response(cls, raw_response: dict[str, Any]) -> EDSLOutput:
        """Parses the API response and returns the response text."""
        return cls.response_handler.parse_response(raw_response)

    async def _async_get_intended_model_call_outcome(
        self,
        user_prompt: str,
        system_prompt: str,
        cache: Cache,
        iteration: int = 0,
        files_list: Optional[List[FileStore]] = None,
        invigilator=None,
    ) -> ModelResponse:
        """Handle caching of responses.

        :param user_prompt: The user's prompt.
        :param system_prompt: The system's prompt.
        :param iteration: The iteration number.
        :param cache: The cache to use.
        :param files_list: The list of files to use.
        :param invigilator: The invigilator to use.

        If the cache isn't being used, it just returns a 'fresh' call to the LLM.
        But if cache is being used, it first checks the database to see if the response is already there.
        If it is, it returns the cached response, but again appends some tracking information.
        If it isn't, it calls the LLM, saves the response to the database, and returns the response with tracking information.

        If self.use_cache is True, then attempts to retrieve the response from the database;
        if not in the DB, calls the LLM and writes the response to the DB.

        >>> from edsl import Cache
        >>> m = LanguageModel.example(test_model = True)
        >>> m._get_intended_model_call_outcome(user_prompt = "Hello", system_prompt = "hello", cache = Cache())
        ModelResponse(...)"""

        if files_list:
            files_hash = "+".join([str(hash(file)) for file in files_list])
            user_prompt_with_hashes = user_prompt + f" {files_hash}"
        else:
            user_prompt_with_hashes = user_prompt

        cache_call_params = {
            "model": str(self.model),
            "parameters": self.parameters,
            "system_prompt": system_prompt,
            "user_prompt": user_prompt_with_hashes,
            "iteration": iteration,
        }
        cached_response, cache_key = cache.fetch(**cache_call_params)

        if cache_used := cached_response is not None:
            response = json.loads(cached_response)
        else:
            f = (
                self.remote_async_execute_model_call
                if hasattr(self, "remote") and self.remote
                else self.async_execute_model_call
            )
            params = {
                "user_prompt": user_prompt,
                "system_prompt": system_prompt,
                "files_list": files_list,
            }
            from edsl.config import CONFIG

            TIMEOUT = float(CONFIG.get("EDSL_API_TIMEOUT"))

            response = await asyncio.wait_for(f(**params), timeout=TIMEOUT)
            new_cache_key = cache.store(
                **cache_call_params, response=response
            )  # store the response in the cache
            assert new_cache_key == cache_key  # should be the same

        cost = self.cost(response)
        return ModelResponse(
            response=response,
            cache_used=cache_used,
            cache_key=cache_key,
            cached_response=cached_response,
            cost=cost,
        )

    _get_intended_model_call_outcome = sync_wrapper(
        _async_get_intended_model_call_outcome
    )

    def simple_ask(
        self,
        question: QuestionBase,
        system_prompt="You are a helpful agent pretending to be a human.",
        top_logprobs=2,
    ):
        """Ask a question and return the response."""
        self.logprobs = True
        self.top_logprobs = top_logprobs
        return self.execute_model_call(
            user_prompt=question.human_readable(), system_prompt=system_prompt
        )

    async def async_get_response(
        self,
        user_prompt: str,
        system_prompt: str,
        cache: Cache,
        iteration: int = 1,
        files_list: Optional[List[FileStore]] = None,
        **kwargs,
    ) -> dict:
        """Get response, parse, and return as string.

        :param user_prompt: The user's prompt.
        :param system_prompt: The system's prompt.
        :param cache: The cache to use.
        :param iteration: The iteration number.
        :param files_list: The list of files to use.

        """
        params = {
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
            "iteration": iteration,
            "cache": cache,
            "files_list": files_list,
        }
        if "invigilator" in kwargs:
            params.update({"invigilator": kwargs["invigilator"]})

        model_inputs = ModelInputs(user_prompt=user_prompt, system_prompt=system_prompt)
        model_outputs: ModelResponse = (
            await self._async_get_intended_model_call_outcome(**params)
        )
        edsl_dict: EDSLOutput = self.parse_response(model_outputs.response)

        agent_response_dict = AgentResponseDict(
            model_inputs=model_inputs,
            model_outputs=model_outputs,
            edsl_dict=edsl_dict,
        )
        return agent_response_dict

    get_response = sync_wrapper(async_get_response)

    def cost(self, raw_response: dict[str, Any]) -> Union[float, str]:
        """Return the dollar cost of a raw response.

        :param raw_response: The raw response from the model.
        """

        usage = self.get_usage_dict(raw_response)
        from edsl.language_models.PriceManager import PriceManager

        price_manger = PriceManager()
        return price_manger.calculate_cost(
            inference_service=self._inference_service_,
            model=self.model,
            usage=usage,
            input_token_name=self.input_token_name,
            output_token_name=self.output_token_name,
        )

    def to_dict(self, add_edsl_version: bool = True) -> dict[str, Any]:
        """Convert instance to a dictionary

        :param add_edsl_version: Whether to add the EDSL version to the dictionary.

        >>> m = LanguageModel.example()
        >>> m.to_dict()
        {'model': '...', 'parameters': {'temperature': ..., 'max_tokens': ..., 'top_p': ..., 'frequency_penalty': ..., 'presence_penalty': ..., 'logprobs': False, 'top_logprobs': ...}, 'edsl_version': '...', 'edsl_class_name': 'LanguageModel'}
        """
        d = {
            "model": self.model,
            "parameters": self.parameters,
        }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = self.__class__.__name__
        return d

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> Type[LanguageModel]:
        """Convert dictionary to a LanguageModel child instance."""
        from edsl.language_models.model import get_model_class

        model_class = get_model_class(data["model"])
        return model_class(**data)

    def __repr__(self) -> str:
        """Return a representation of the object."""
        param_string = ", ".join(
            f"{key} = {value}" for key, value in self.parameters.items()
        )
        return (
            f"Model(model_name = '{self.model}'"
            + (f", {param_string}" if param_string else "")
            + ")"
        )

    def __add__(self, other_model: Type[LanguageModel]) -> Type[LanguageModel]:
        """Combine two models into a single model (other_model takes precedence over self)."""
        import warnings

        warnings.warn(
            f"""Warning: one model is replacing another. If you want to run both models, use a single `by` e.g., 
              by(m1, m2, m3) not by(m1).by(m2).by(m3)."""
        )
        return other_model or self

    @classmethod
    def example(
        cls,
        test_model: bool = False,
        canned_response: str = "Hello world",
        throw_exception: bool = False,
    ) -> LanguageModel:
        """Return a default instance of the class.

        >>> from edsl.language_models import LanguageModel
        >>> m = LanguageModel.example(test_model = True, canned_response = "WOWZA!")
        >>> isinstance(m, LanguageModel)
        True
        >>> from edsl import QuestionFreeText
        >>> q = QuestionFreeText(question_text = "What is your name?", question_name = 'example')
        >>> q.by(m).run(cache = False, disable_remote_cache = True, disable_remote_inference = True).select('example').first()
        'WOWZA!'
        >>> m = LanguageModel.example(test_model = True, canned_response = "WOWZA!", throw_exception = True)
        >>> r = q.by(m).run(cache = False, disable_remote_cache = True, disable_remote_inference = True, print_exceptions = True)
        Exception report saved to ...
        Also see: ...
        """
        from edsl.language_models.model import Model

        if test_model:
            m = Model(
                "test", canned_response=canned_response, throw_exception=throw_exception
            )
            return m
        else:
            return Model(skip_api_key_check=True)

    def from_cache(self, cache: "Cache") -> LanguageModel:

        from copy import deepcopy
        from types import MethodType
        from edsl import Cache

        new_instance = deepcopy(self)
        print("Cache entries", len(cache))
        new_instance.cache = Cache(
            data={k: v for k, v in cache.items() if v.model == self.model}
        )
        print("Cache entries with same model", len(new_instance.cache))

        new_instance.user_prompts = [
            ce.user_prompt for ce in new_instance.cache.values()
        ]
        new_instance.system_prompts = [
            ce.system_prompt for ce in new_instance.cache.values()
        ]

        async def async_execute_model_call(self, user_prompt: str, system_prompt: str):
            cache_call_params = {
                "model": str(self.model),
                "parameters": self.parameters,
                "system_prompt": system_prompt,
                "user_prompt": user_prompt,
                "iteration": 1,
            }
            cached_response, cache_key = cache.fetch(**cache_call_params)
            response = json.loads(cached_response)
            cost = 0
            return ModelResponse(
                response=response,
                cache_used=True,
                cache_key=cache_key,
                cached_response=cached_response,
                cost=cost,
            )

        # Bind the new method to the copied instance
        setattr(
            new_instance,
            "async_execute_model_call",
            MethodType(async_execute_model_call, new_instance),
        )

        return new_instance


if __name__ == "__main__":
    """Run the module's test suite."""
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.questions.QuestionBase import QuestionBase

from edsl.surveys.Rule import Rule
from .base import RulePriority, EndOfSurvey
from edsl.exceptions.surveys import SurveyError, SurveyCreationError


class ValidatedString(str):
    def __new__(cls, content):
        if "<>" in content:
            raise SurveyCreationError(
                "The expression contains '<>', which is not allowed. You probably mean '!='."
            )
        return super().__new__(cls, content)


class RuleManager:
    def __init__(self, survey):
        self.survey = survey

    def _get_question_index(
        self, q: Union["QuestionBase", str, EndOfSurvey.__class__]
    ) -> Union[int, EndOfSurvey.__class__]:
        """Return the index of the question or EndOfSurvey object.

        :param q: The question or question name to get the index of.

        It can handle it if the user passes in the question name, the question object, or the EndOfSurvey object.

        >>> from edsl.questions import QuestionFreeText
        >>> from edsl import Survey
        >>> s = Survey.example()
        >>> s._get_question_index("q0")
        0

        This doesnt' work with questions that don't exist:

        >>> s._get_question_index("poop")
        Traceback (most recent call last):
        ...
        edsl.exceptions.surveys.SurveyError: Question name poop not found in survey. The current question names are {'q0': 0, 'q1': 1, 'q2': 2}.
        ...
        """
        if q == EndOfSurvey:
            return EndOfSurvey
        else:
            question_name = q if isinstance(q, str) else q.question_name
            if question_name not in self.survey.question_name_to_index:
                raise SurveyError(
                    f"""Question name {question_name} not found in survey. The current question names are {self.survey.question_name_to_index}."""
                )
            return self.survey.question_name_to_index[question_name]

    def _get_new_rule_priority(
        self, question_index: int, before_rule: bool = False
    ) -> int:
        """Return the priority for the new rule.

        :param question_index: The index of the question to add the rule to.
        :param before_rule: Whether the rule is evaluated before the question is answered.

        >>> from edsl import Survey
        >>> s = Survey.example()
        >>> RuleManager(s)._get_new_rule_priority(0)
        1
        """
        current_priorities = [
            rule.priority
            for rule in self.survey.rule_collection.applicable_rules(
                question_index, before_rule
            )
        ]
        if len(current_priorities) == 0:
            return RulePriority.DEFAULT.value + 1

        max_priority = max(current_priorities)
        # newer rules take priority over older rules
        new_priority = (
            RulePriority.DEFAULT.value
            if len(current_priorities) == 0
            else max_priority + 1
        )
        return new_priority

    def add_rule(
        self,
        question: Union["QuestionBase", str],
        expression: str,
        next_question: Union["QuestionBase", str, int],
        before_rule: bool = False,
    ) -> "Survey":
        """
        Add a rule to a Question of the Survey with the appropriate priority.

        :param question: The question to add the rule to.
        :param expression: The expression to evaluate.
        :param next_question: The next question to go to if the rule is true.
        :param before_rule: Whether the rule is evaluated before the question is answered.


        - The last rule added for the question will have the highest priority.
        - If there are no rules, the rule added gets priority -1.
        """
        question_index = self.survey._get_question_index(question)  # Fix

        # Might not have the name of the next question yet
        if isinstance(next_question, int):
            next_question_index = next_question
        else:
            next_question_index = self._get_question_index(next_question)

        new_priority = self._get_new_rule_priority(question_index, before_rule)  # fix

        self.survey.rule_collection.add_rule(
            Rule(
                current_q=question_index,
                expression=expression,
                next_q=next_question_index,
                question_name_to_index=self.survey.question_name_to_index,
                priority=new_priority,
                before_rule=before_rule,
            )
        )

        return self.survey

    def add_stop_rule(
        self, question: Union["QuestionBase", str], expression: str
    ) -> "Survey":
        """Add a rule that stops the survey.
        The rule is evaluated *after* the question is answered. If the rule is true, the survey ends.

        :param question: The question to add the stop rule to.
        :param expression: The expression to evaluate.

        If this rule is true, the survey ends.

        Here, answering "yes" to q0 ends the survey:

        >>> from edsl import Survey
        >>> s = Survey.example().add_stop_rule("q0", "q0 == 'yes'")
        >>> s.next_question("q0", {"q0": "yes"})
        EndOfSurvey

        By comparison, answering "no" to q0 does not end the survey:

        >>> s.next_question("q0", {"q0": "no"}).question_name
        'q1'

        >>> s.add_stop_rule("q0", "q1 <> 'yes'")
        Traceback (most recent call last):
        ...
        edsl.exceptions.surveys.SurveyCreationError: The expression contains '<>', which is not allowed. You probably mean '!='.
        ...
        """
        expression = ValidatedString(expression)
        prior_question_appears = False
        for prior_question in self.survey.questions:
            if prior_question.question_name in expression:
                prior_question_appears = True

        if not prior_question_appears:
            import warnings

            warnings.warn(
                f"The expression {expression} does not contain any prior question names. This is probably a mistake."
            )
        self.survey.add_rule(question, expression, EndOfSurvey)
        return self.survey



from textwrap import dedent


class JobsErrors(Exception):
    pass


class JobsRunError(JobsErrors):
    pass


class MissingRemoteInferenceError(JobsErrors):
    def __init__(self):
        message = dedent(
            """\\
                        You are trying to run the job remotely, but you have not set the EXPECTED_PARROT_INFERENCE_URL environment variable.
                        This remote running service is not quite ready yet! 
                        But please see https://docs.expectedparrot.com/en/latest/coop.html for what we are working on.
                        """
        )
        super().__init__(message)


class InterviewError(Exception):
    pass


class InterviewErrorPriorTaskCanceled(InterviewError):
    pass


class InterviewTimeoutError(InterviewError):
    pass



"""A survey has a memory plan that specifies what the agent should remember when answering a question."""

from collections import UserDict, defaultdict
from typing import Optional

# from edsl.surveys.Memory import Memory
# from edsl.prompts.Prompt import Prompt
# from edsl.surveys.DAG import DAG


class MemoryPlan(UserDict):
    """A survey has a memory plan that specifies what the agent should remember when answering a question.

    The basic structure of a memory plan is a dictionary of focal questions to memories.

    {focal_question1: [prior_question1, prior_question2, ...], focal_question: [prior_question3]}
    """

    def __init__(self, survey: Optional["Survey"] = None, data: Optional[dict] = None):
        """Initialize a memory plan.

        The actual 'data' attributes of the memory plan are a dictionary of focal questions to memories.
        """
        if survey is not None:
            self.survey = survey
            self.survey_question_names = [q.question_name for q in survey.questions]
            self.question_texts = [q.question_text for q in survey.questions]
        super().__init__(data or {})

    @property
    def name_to_text(self) -> dict:
        """Return a dictionary mapping question names to question texts."""
        return dict(zip(self.survey_question_names, self.question_texts))

    def add_question(self, question: "QuestionBase") -> None:
        """Add a question to the survey.

        :param question: A question to add to the survey

        """
        self.survey_question_names.append(question.question_name)
        self.question_texts.append(question.question_text)

    def _check_valid_question_name(self, question_name: str) -> None:
        """Ensure a passed question name is valid.

        :param question_name: The name of the question to check.

        """
        if question_name not in self.survey_question_names:
            raise ValueError(
                f"{question_name} is not in the survey. Current names are {self.survey_question_names}"
            )

    def get_memory_prompt_fragment(
        self, focal_question: str, answers: dict
    ) -> "Prompt":
        """Generate the prompt fragment descripting that past question and answer.

        :param focal_question: The current question being answered.
        :param answers: A dictionary of question names to answers.

        """
        from edsl.prompts.Prompt import Prompt

        self._check_valid_question_name(focal_question)

        if focal_question not in self:
            return Prompt("")

        q_and_a_pairs = [
            (self.name_to_text[question_name], answers.get(question_name, None))
            for question_name in self[focal_question]
        ]

        base_prompt_text = """
        Before the question you are now answering, you already answered the following question(s):
        """

        def gen_line(question_text, answer):
            """Return a line of memory."""
            return f"\tQuestion: {question_text}\n\tAnswer: {answer}\n"

        lines = [gen_line(*pair) for pair in q_and_a_pairs]
        if lines:
            return Prompt(
                base_prompt_text + "\n Prior questions and answers:".join(lines)
            )
        else:
            return Prompt("")

    def _check_order(self, focal_question: str, prior_question: str) -> None:
        """Ensure the prior question comes before the focal question."""
        focal_index = self.survey_question_names.index(focal_question)
        prior_index = self.survey_question_names.index(prior_question)
        if focal_index <= prior_index:
            raise ValueError(f"{prior_question} must come before {focal_question}.")

    def add_single_memory(self, focal_question: str, prior_question: str) -> None:
        """Add a single memory to the memory plan.

        :param focal_question: The current question being answered.
        :param prior_question: The question that was answered before the focal question that should be remembered.

        >>> mp = MemoryPlan.example()
        >>> mp.add_single_memory("q0", "q1")
        Traceback (most recent call last):
        ...
        ValueError: q1 must come before q0.

        >>> mp = MemoryPlan.example()
        >>> mp.add_single_memory("q0", "crap")
        Traceback (most recent call last):
        ...
        ValueError: crap is not in the survey. Current names are ['q0', 'q1', 'q2']

        >>> mp = MemoryPlan.example()
        >>> mp.add_single_memory("crap", "q0")
        Traceback (most recent call last):
        ...
        ValueError: crap is not in the survey. Current names are ['q0', 'q1', 'q2']
        """
        self._check_valid_question_name(focal_question)
        self._check_valid_question_name(prior_question)
        self._check_order(focal_question, prior_question)
        from edsl.surveys.Memory import Memory

        if focal_question not in self:
            memory = Memory()
            memory.add_prior_question(prior_question)
            self[focal_question] = memory
        else:
            self[focal_question].add_prior_question(prior_question)

    def add_memory_collection(
        self, focal_question: str, prior_questions: list[str]
    ) -> None:
        """Add a collection of prior questions to the memory plan.

        :param focal_question: The current question being answered.
        :param prior_questions: A list of questions that were answered before the focal question that should be remembered.
        """
        for question in prior_questions:
            self.add_single_memory(focal_question, question)

    def to_dict(self, add_edsl_version=True) -> dict:
        """Serialize the memory plan to a dictionary.

        >>> mp = MemoryPlan.example()
        >>> mp.to_dict()
        {'survey_question_names': ['q0', 'q1', 'q2'], 'survey_question_texts': ['Do you like school?', 'Why not?', 'Why?'], 'data': {'q1': {'prior_questions': ['q0']}}}
        """
        newdata = {}
        for question_name, memory in self.items():
            newdata[question_name] = memory.to_dict()

        return {
            "survey_question_names": self.survey_question_names,
            "survey_question_texts": self.question_texts,
            "data": newdata,
        }

    @classmethod
    def from_dict(cls, data) -> "MemoryPlan":
        """Deserialize a memory plan from a dictionary."""
        from edsl.surveys.Memory import Memory

        newdata = {}
        for question_name, memory in data["data"].items():
            newdata[question_name] = Memory.from_dict(memory)

        memory_plan = cls(survey=None, data=newdata)
        memory_plan.survey_question_names = data["survey_question_names"]
        memory_plan.question_texts = data["survey_question_texts"]
        return memory_plan

    def _indexify(self, d: dict):
        """Convert a dictionary of question names to a dictionary of question indices.

        :param d: A dictionary of question names to indices.
        """
        new_d = {}
        for k, v in d.items():
            key = self.survey_question_names.index(k)
            new_v = set({self.survey_question_names.index(q) for q in v})
            new_d[key] = new_v
        return new_d

    @property
    def dag(self) -> "DAG":
        """Return a directed acyclic graph of the memory plan.

        >>> mp = MemoryPlan.example()
        >>> mp.dag
        {1: {0}}
        """
        from edsl.surveys.DAG import DAG

        d = defaultdict(set)
        for focal_question, memory in self.items():
            for prior_question in memory:
                d[focal_question].add(prior_question)
        return DAG(self._indexify(d))

    @classmethod
    def example(cls):
        """Return an example memory plan."""
        from edsl import Survey

        mp = cls(survey=Survey.example())
        mp.add_single_memory("q1", "q0")
        return mp

    def remove_question(self, question_name: str) -> None:
        """Remove a question from the memory plan.

        :param question_name: The name of the question to remove.
        """
        self._check_valid_question_name(question_name)

        # Remove the question from survey_question_names and question_texts
        index = self.survey_question_names.index(question_name)
        self.survey_question_names.pop(index)
        self.question_texts.pop(index)

        # Remove the question from the memory plan if it's a focal question
        self.pop(question_name, None)

        # Remove the question from all memories where it appears as a prior question
        for focal_question, memory in self.items():
            memory.remove_prior_question(question_name)

        # Update the DAG
        self.dag.remove_node(index)

    def remove_prior_question(self, question_name: str) -> None:
        """Remove a prior question from the memory."""
        self.prior_questions = [q for q in self.prior_questions if q != question_name]


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.surveys.base import EndOfSurvey
from edsl.surveys.DAG import DAG
from edsl.exceptions.surveys import SurveyError


class ConstructDAG:
    def __init__(self, survey):
        self.survey = survey
        self.questions = survey.questions

        self.parameters_by_question = self.survey.parameters_by_question
        self.question_name_to_index = self.survey.question_name_to_index

    def dag(self, textify: bool = False) -> DAG:
        memory_dag = self.survey.memory_plan.dag
        rule_dag = self.survey.rule_collection.dag
        piping_dag = self.piping_dag
        if textify:
            memory_dag = DAG(self.textify(memory_dag))
            rule_dag = DAG(self.textify(rule_dag))
            piping_dag = DAG(self.textify(piping_dag))
        return memory_dag + rule_dag + piping_dag

    @property
    def piping_dag(self) -> DAG:
        """Figures out the DAG of piping dependencies.

        >>> from edsl import Survey
        >>> from edsl import QuestionFreeText
        >>> q0 = QuestionFreeText(question_text="Here is a question", question_name="q0")
        >>> q1 = QuestionFreeText(question_text="You previously answered {{ q0 }}---how do you feel now?", question_name="q1")
        >>> s = Survey([q0, q1])
        >>> ConstructDAG(s).piping_dag
        {1: {0}}
        """
        d = {}
        for question_name, depenencies in self.parameters_by_question.items():
            if depenencies:
                question_index = self.question_name_to_index[question_name]
                for dependency in depenencies:
                    if dependency not in self.question_name_to_index:
                        pass
                    else:
                        dependency_index = self.question_name_to_index[dependency]
                        if question_index not in d:
                            d[question_index] = set()
                        d[question_index].add(dependency_index)
        return d

    def textify(self, index_dag: DAG) -> DAG:
        """Convert the DAG of question indices to a DAG of question names.

        :param index_dag: The DAG of question indices.

        Example:

        >>> from edsl import Survey
        >>> s = Survey.example()
        >>> d = s.dag()
        >>> d
        {1: {0}, 2: {0}}
        >>> ConstructDAG(s).textify(d)
        {'q1': {'q0'}, 'q2': {'q0'}}
        """

        def get_name(index: int):
            """Return the name of the question given the index."""
            if index >= len(self.questions):
                return EndOfSurvey
            try:
                return self.questions[index].question_name
            except IndexError:
                print(
                    f"The index is {index} but the length of the questions is {len(self.questions)}"
                )
                raise SurveyError

        try:
            text_dag = {}
            for child_index, parent_indices in index_dag.items():
                parent_names = {get_name(index) for index in parent_indices}
                child_name = get_name(child_index)
                text_dag[child_name] = parent_names
            return text_dag
        except IndexError:
            raise


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations
import copy
import itertools
from typing import Optional, List, Callable, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.questions.QuestionBase import QuestionBase
    from edsl.scenarios.ScenarioList import ScenarioList


class QuestionBaseGenMixin:
    """Mixin for QuestionBase."""

    def copy(self) -> QuestionBase:
        """Return a deep copy of the question.

        >>> from edsl.questions import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite color?")
        >>> q2 = q.copy()
        >>> q2.question_name
        'color'

        """
        return copy.deepcopy(self)

    def option_permutations(self) -> list[QuestionBase]:
        """Return a list of questions with all possible permutations of the options.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice as Q
        >>> len(Q.example().option_permutations())
        24
        """

        if not hasattr(self, "question_options"):
            return [self]

        questions = []
        for index, permutation in enumerate(
            itertools.permutations(self.question_options)
        ):
            question = copy.deepcopy(self)
            question.question_options = list(permutation)
            question.question_name = f"{self.question_name}_{index}"
            questions.append(question)
        return questions

    def draw(self) -> "QuestionBase":
        """Return a new question with a randomly selected permutation of the options.

        If the question has no options, returns a copy of the original question.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice as Q
        >>> q = Q.example()
        >>> drawn = q.draw()
        >>> len(drawn.question_options) == len(q.question_options)
        True
        >>> q is drawn
        False
        """

        if not hasattr(self, "question_options"):
            return copy.deepcopy(self)

        import random

        question = copy.deepcopy(self)
        question.question_options = list(
            random.sample(self.question_options, len(self.question_options))
        )
        return question

    def loop(self, scenario_list: ScenarioList) -> List[QuestionBase]:
        """Return a list of questions with the question name modified for each scenario.

        :param scenario_list: The list of scenarios to loop through.

        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.scenarios.ScenarioList import ScenarioList
        >>> q = QuestionFreeText(question_text = "What are your thoughts on: {{ subject}}?", question_name = "base_{{subject}}")
        >>> len(q.loop(ScenarioList.from_list("subject", ["Math", "Economics", "Chemistry"])))
        3
        """
        from edsl.questions.loop_processor import LoopProcessor

        lp = LoopProcessor(self)
        return lp.process_templates(scenario_list)

    def render(self, replacement_dict: dict) -> "QuestionBase":
        """Render the question components as jinja2 templates with the replacement dictionary.

        :param replacement_dict: The dictionary of values to replace in the question components.

        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite {{ thing }}?")
        >>> q.render({"thing": "color"})
        Question('free_text', question_name = \"""color\""", question_text = \"""What is your favorite color?\""")

        """
        from jinja2 import Environment
        from edsl.scenarios.Scenario import Scenario

        strings_only_replacement_dict = {
            k: v for k, v in replacement_dict.items() if not isinstance(v, Scenario)
        }

        def render_string(value: str) -> str:
            if value is None or not isinstance(value, str):
                return value
            else:
                try:
                    return (
                        Environment()
                        .from_string(value)
                        .render(strings_only_replacement_dict)
                    )
                except Exception as e:
                    import warnings

                    warnings.warn("Failed to render string: " + value)
                    # breakpoint()
                    return value

        return self.apply_function(render_string)

    def apply_function(
        self, func: Callable, exclude_components: List[str] = None
    ) -> QuestionBase:
        """Apply a function to the question parts

        :param func: The function to apply to the question parts.
        :param exclude_components: The components to exclude from the function application.

        >>> from edsl.questions import QuestionFreeText
        >>> q = QuestionFreeText(question_name = "color", question_text = "What is your favorite color?")
        >>> shouting = lambda x: x.upper()
        >>> q.apply_function(shouting)
        Question('free_text', question_name = \"""color\""", question_text = \"""WHAT IS YOUR FAVORITE COLOR?\""")

        >>> q.apply_function(shouting, exclude_components = ["question_type"])
        Question('free_text', question_name = \"""COLOR\""", question_text = \"""WHAT IS YOUR FAVORITE COLOR?\""")

        """
        from edsl.questions.QuestionBase import QuestionBase

        if exclude_components is None:
            exclude_components = ["question_name", "question_type"]

        d = copy.deepcopy(self.to_dict(add_edsl_version=False))
        for key, value in d.items():
            if key in exclude_components:
                continue
            if isinstance(value, dict):
                for k, v in value.items():
                    value[k] = func(v)
                d[key] = value
                continue
            if isinstance(value, list):
                value = [func(v) for v in value]
                d[key] = value
                continue
            d[key] = func(value)
        return QuestionBase.from_dict(d)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations
from typing import Any, List, Union, Dict, Optional
from pathlib import Path

# from jinja2 import Undefined


from edsl.exceptions.prompts import TemplateRenderError
from edsl.Base import PersistenceMixin, RepresentationMixin

MAX_NESTING = 100


class Prompt(PersistenceMixin, RepresentationMixin):
    """Class for creating a prompt to be used in a survey."""

    default_instructions: Optional[str] = "Do good things, friendly LLM!"

    def __len__(self):
        """Return the length of the prompt text."""
        return len(self.text)

    @classmethod
    def prompt_attributes(cls) -> List[str]:
        """Return the prompt class attributes."""
        return {k: v for k, v in cls.__dict__.items() if not k.startswith("_")}

    def __init__(self, text: Optional[str] = None):
        """Create a `Prompt` object.

        :param text: The text of the prompt.
        """
        if text is None:
            if hasattr(self, "default_instructions"):
                text = self.default_instructions
            else:
                text = ""
        if isinstance(text, Prompt):
            # make it idempotent w/ a prompt
            text = text.text
        self._text = text

    @classmethod
    def from_txt(cls, filename: str) -> PromptBase:
        """Create a `Prompt` from text.

        :param text: The text of the prompt.
        """
        with open(filename, "r") as f:
            text = f.read()
        return cls(text=text)

    @classmethod
    def from_template(
        cls,
        file_name: str,
        path_to_folder: Optional[Union[str, Path]] = None,
        **kwargs: Dict[str, Any],
    ) -> "PromptBase":
        """Create a `PromptBase` from a Jinja template.

        Args:
            file_name (str): The name of the Jinja template file.
            path_to_folder (Union[str, Path]): The path to the folder containing the template.
                            Can be absolute or relative.
            **kwargs: Variables to be passed to the template for rendering.

        Returns:
            PromptBase: An instance of PromptBase with the rendered template as text.
        """
        # if file_name lacks the .j2 extension, add it
        if not file_name.endswith(".jinja"):
            file_name += ".jinja"

        # Convert path_to_folder to a Path object if it's a string
        if path_to_folder is None:
            from importlib import resources
            import os

            path_to_folder = resources.path("edsl.questions", "prompt_templates")

        try:
            folder_path = Path(path_to_folder)
        except Exception as e:
            raise ValueError(f"Invalid path: {path_to_folder}. Error: {e}")

        with open(folder_path.joinpath(file_name), "r") as f:
            text = f.read()
        return cls(text=text)

    @property
    def text(self):
        """Return the `Prompt` text."""
        return self._text

    def __add__(self, other_prompt):
        """Add two prompts together.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt("How are you?")
        >>> p + p2
        Prompt(text=\"""Hello, {{person}}How are you?\""")

        >>> p + "How are you?"
        Prompt(text=\"""Hello, {{person}}How are you?\""")
        """
        if isinstance(other_prompt, str):
            return self.__class__(self.text + other_prompt)
        else:
            return self.__class__(text=self.text + other_prompt.text)

    def __str__(self):
        """Return the `Prompt` text.

        Example:
        >>> p = Prompt("Hello, {{person}}")
        >>> str(p)
        'Hello, {{person}}'
        """
        return self.text

    def __contains__(self, text_to_check):
        """Check if the text_to_check is in the `Prompt` text.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> "person" in p
        True
        >>> "person2" in p
        False
        """
        return text_to_check in self.text

    def __repr__(self):
        """Return the `Prompt` text.

        Example:
        >>> p = Prompt("Hello, {{person}}")
        >>> p
        Prompt(text=\"""Hello, {{person}}\""")
        """
        return f'Prompt(text="""{self.text}""")'

    def template_variables(self) -> list[str]:
        """Return the the variables in the template.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.template_variables()
        ['person']

        """
        return self._template_variables(self.text)

    @staticmethod
    def _template_variables(template: str) -> list[str]:
        """Find and return the template variables.

        :param template: The template to find the variables in.

        """
        from jinja2 import Environment, meta, Undefined

        class PreserveUndefined(Undefined):
            def __str__(self):
                return "{{ " + str(self._undefined_name) + " }}"

        env = Environment(undefined=PreserveUndefined)
        ast = env.parse(template)
        return list(meta.find_undeclared_variables(ast))

    def undefined_template_variables(self, replacement_dict: dict):
        """Return the variables in the template that are not in the replacement_dict.

        :param replacement_dict: A dictionary of replacements to populate the template.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.undefined_template_variables({"person": "John"})
        []

        >>> p = Prompt("Hello, {{title}} {{person}}")
        >>> p.undefined_template_variables({"person": "John"})
        ['title']
        """
        return [var for var in self.template_variables() if var not in replacement_dict]

    def unused_traits(self, traits: dict):
        """Return the traits that are not used in the template."""
        return [trait for trait in traits if trait not in self.template_variables()]

    @property
    def has_variables(self) -> bool:
        """Return True if the prompt has variables.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.has_variables
        True

        >>> p = Prompt("Hello, person")
        >>> p.has_variables
        False
        """
        return len(self.template_variables()) > 0

    def render(self, primary_replacement: dict, **additional_replacements) -> str:
        """Render the prompt with the replacements.

        :param primary_replacement: The primary replacement dictionary.
        :param additional_replacements: Additional replacement dictionaries.

        >>> p = Prompt("Hello, {{person}}")
        >>> p.render({"person": "John"})
        Prompt(text=\"""Hello, John\""")

        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Horton"})
        Prompt(text=\"""Hello, Mr. Horton\""")

        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Ho{{letter}}ton"}, max_nesting = 1)
        Prompt(text=\"""Hello, Mr. Ho{{ letter }}ton\""")

        >>> p.render({"person": "Mr. {{last_name}}"})
        Prompt(text=\"""Hello, Mr. {{ last_name }}\""")
        """
        try:
            new_text = self._render(
                self.text, primary_replacement, **additional_replacements
            )
            return self.__class__(text=new_text)
        except Exception as e:
            print(f"Error rendering prompt: {e}")
            return self

    @staticmethod
    def _render(
        text: str, primary_replacement, **additional_replacements
    ) -> "PromptBase":
        """Render the template text with variables replaced from the provided named dictionaries.

        :param text: The text to render.
        :param primary_replacement: The primary replacement dictionary.
        :param additional_replacements: Additional replacement dictionaries.

        Allows for nested variable resolution up to a specified maximum nesting depth.

        Example:

        >>> codebook = {"age": "Age"}
        >>> p = Prompt("You are an agent named {{ name }}. {{ codebook['age']}}: {{ age }}")
        >>> p.render({"name": "John", "age": 44}, codebook=codebook)
        Prompt(text=\"""You are an agent named John. Age: 44\""")
        """
        from jinja2 import Environment, meta, TemplateSyntaxError, Undefined

        class PreserveUndefined(Undefined):
            def __str__(self):
                return "{{ " + str(self._undefined_name) + " }}"

        env = Environment(undefined=PreserveUndefined)
        try:
            previous_text = None
            for _ in range(MAX_NESTING):
                # breakpoint()
                rendered_text = env.from_string(text).render(
                    primary_replacement, **additional_replacements
                )
                if rendered_text == previous_text:
                    # No more changes, so return the rendered text
                    return rendered_text
                previous_text = text
                text = rendered_text

            # If the loop exits without returning, it indicates too much nesting
            raise TemplateRenderError(
                "Too much nesting - you created an infinite loop here, pal"
            )
        except TemplateSyntaxError as e:
            raise TemplateRenderError(
                f"Template syntax error: {e}. Bad template: {text}"
            )

    def to_dict(self, add_edsl_version=False) -> dict[str, Any]:
        """Return the `Prompt` as a dictionary.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.to_dict()
        {'text': 'Hello, {{person}}', 'class_name': 'Prompt'}

        """
        return {"text": self.text, "class_name": self.__class__.__name__}

    @classmethod
    def from_dict(cls, data) -> PromptBase:
        """Create a `Prompt` from a dictionary.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt.from_dict(p.to_dict())
        >>> p2
        Prompt(text=\"""Hello, {{person}}\""")

        """
        # class_name = data["class_name"]
        return Prompt(text=data["text"])

    # def rich_print(self):
    #     """Display an object as a table."""
    #     table = Table(title="Prompt")
    #     table.add_column("Attribute", style="bold")
    #     table.add_column("Value")

    #     to_display = self.__dict__.copy()
    #     for attr_name, attr_value in to_display.items():
    #         table.add_row(attr_name, repr(attr_value))
    #     table.add_row("Component type", str(self.component_type))
    #     table.add_row("Model", str(getattr(self, "model", "Not specified")))
    #     return table

    @classmethod
    def example(cls):
        """Return an example of the prompt."""
        return cls(cls.default_instructions)


if __name__ == "__main__":
    print("Running doctests...")
    import doctest

    doctest.testmod()

# from edsl.prompts.library.question_multiple_choice import *
# from edsl.prompts.library.agent_instructions import *
# from edsl.prompts.library.agent_persona import *

# from edsl.prompts.library.question_budget import *
# from edsl.prompts.library.question_checkbox import *
# from edsl.prompts.library.question_freetext import *
# from edsl.prompts.library.question_linear_scale import *
# from edsl.prompts.library.question_numerical import *
# from edsl.prompts.library.question_rank import *
# from edsl.prompts.library.question_extract import *
# from edsl.prompts.library.question_list import *



from dataclasses import dataclass, asdict
from typing import Optional


@dataclass
class APIKeyEntry:
    """A class representing an API key entry.

    >>> entry = APIKeyEntry.example()
    >>> entry.service
    'openai'
    >>> entry.name
    'OPENAI_API_KEY'
    >>> entry.value
    'sk-abcd1234'
    >>> entry.source
    'env'
    """

    service: str
    name: str
    value: str
    source: Optional[str] = None

    @classmethod
    def example(cls):
        return APIKeyEntry(
            service="openai", name="OPENAI_API_KEY", value="sk-abcd1234", source="env"
        )


@dataclass
class LimitEntry:
    """A class representing rate limit entries for a service.

    >>> limit = LimitEntry.example()
    >>> limit.service
    'openai'
    >>> limit.rpm
    60
    >>> limit.tpm
    100000
    >>> limit.source
    'config'
    """

    service: str
    rpm: int
    tpm: int
    source: Optional[str] = None

    @classmethod
    def example(cls):
        return LimitEntry(service="openai", rpm=60, tpm=100000, source="config")


@dataclass
class APIIDEntry:
    """A class representing an API ID entry.

    >>> id_entry = APIIDEntry.example()
    >>> id_entry.service
    'bedrock'
    >>> id_entry.name
    'AWS_ACCESS_KEY_ID'
    >>> id_entry.value
    'AKIA1234'
    >>> id_entry.source
    'env'
    """

    service: str
    name: str
    value: str
    source: Optional[str] = None

    @classmethod
    def example(cls):
        return APIIDEntry(
            service="bedrock", name="AWS_ACCESS_KEY_ID", value="AKIA1234", source="env"
        )


@dataclass
class LanguageModelInput:
    """A class representing input configuration for a language model service.

    >>> lm_input = LanguageModelInput.example()
    >>> lm_input.api_token
    'sk-abcd123'
    >>> lm_input.rpm
    60
    >>> lm_input.tpm
    100000
    >>> lm_input.api_id


    Test dictionary conversion:
    >>> d = lm_input.to_dict()
    >>> isinstance(d, dict)
    True
    >>> LanguageModelInput.from_dict(d).api_token == lm_input.api_token
    True
    """

    api_token: str
    rpm: int
    tpm: int
    api_id: Optional[str] = None
    token_source: Optional[str] = None
    limit_source: Optional[str] = None
    id_source: Optional[str] = None

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    @classmethod
    def example(cls):
        return LanguageModelInput(
            api_token="sk-abcd123", tpm=100000, rpm=60, api_id=None
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import textwrap
from random import random
from typing import Optional, TYPE_CHECKING, List

from edsl.utilities.PrettyList import PrettyList
from edsl.config import CONFIG

from edsl.inference_services.InferenceServicesCollection import (
    InferenceServicesCollection,
)
from edsl.inference_services.data_structures import AvailableModels
from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.enums import InferenceServiceLiteral

if TYPE_CHECKING:
    from edsl.results.Dataset import Dataset


def get_model_class(model_name, registry: Optional[InferenceServicesCollection] = None):
    from edsl.inference_services.registry import default

    registry = registry or default
    factory = registry.create_model_factory(model_name)
    return factory


class Meta(type):
    def __repr__(cls):
        return textwrap.dedent(
            f"""\
        Available models: {cls.available()}
        
        To create an instance, you can do: 
        >>> m = Model('gpt-4-1106-preview', temperature=0.5, ...)
        
        To get the default model, you can leave out the model name. 
        To see the available models, you can do:
        >>> Model.available()
        """
        )


class Model(metaclass=Meta):
    default_model = CONFIG.get("EDSL_DEFAULT_MODEL")
    _registry: InferenceServicesCollection = None  # Class-level registry storage

    @classmethod
    def get_registry(cls) -> InferenceServicesCollection:
        """Get the current registry or initialize with default if None"""
        if cls._registry is None:
            from edsl.inference_services.registry import default

            cls._registry = default
        return cls._registry

    @classmethod
    def set_registry(cls, registry: InferenceServicesCollection) -> None:
        """Set a new registry"""
        cls._registry = registry

    def __new__(
        cls,
        model_name: Optional[str] = None,
        service_name: Optional[InferenceServiceLiteral] = None,
        registry: Optional[InferenceServicesCollection] = None,
        *args,
        **kwargs,
    ):
        "Instantiate a new language model."
        # Map index to the respective subclass
        if model_name is None:
            model_name = (
                cls.default_model
            )  # when model_name is None, use the default model, set in the config file

        if registry is not None:
            cls.set_registry(registry)

        if isinstance(model_name, int):  # can refer to a model by index
            model_name = cls.available(name_only=True)[model_name]

        factory = cls.get_registry().create_model_factory(
            model_name, service_name=service_name
        )
        return factory(*args, **kwargs)

    @classmethod
    def add_model(cls, service_name, model_name) -> None:
        cls.get_registry().add_model(service_name, model_name)

    @classmethod
    def service_classes(cls) -> List["InferenceServiceABC"]:
        """Returns a list of service classes.

        >>> Model.service_classes()
        [...]
        """
        return [r for r in cls.services(name_only=True)]

    @classmethod
    def services(cls, name_only: bool = False) -> List[str]:
        """Returns a list of services, annotated with whether the user has local keys for them."""
        services_with_local_keys = set(cls.key_info().select("service").to_list())
        f = lambda service_name: (
            "yes" if service_name in services_with_local_keys else " "
        )
        if name_only:
            return PrettyList(
                [r._inference_service_ for r in cls.get_registry().services],
                columns=["Service Name"],
            )
        else:
            return PrettyList(
                [
                    (r._inference_service_, f(r._inference_service_))
                    for r in cls.get_registry().services
                ],
                columns=["Service Name", "Local key?"],
            )

    @classmethod
    def services_with_local_keys(cls) -> set:
        """Returns a list of services for which the user has local keys."""
        return set(cls.key_info().select("service").to_list())

    @classmethod
    def key_info(cls, obscure_api_key: bool = True) -> "Dataset":
        """Returns a dataset of local key information."""
        from edsl.language_models.key_management.KeyLookupCollection import (
            KeyLookupCollection,
        )
        from edsl.scenarios import Scenario, ScenarioList

        klc = KeyLookupCollection()
        klc.add_key_lookup(fetch_order=None)
        sl = ScenarioList()
        for service, entry in list(klc.data.values())[0].items():
            sl.append(Scenario({"service": service} | entry.to_dict()))
        if obscure_api_key:
            for service in sl:
                service["api_token"] = (
                    service["api_token"][:4] + "..." + service["api_token"][-4:]
                )
        return sl.to_dataset()

    @classmethod
    def search_models(cls, search_term: str):
        return cls.available(search_term=search_term)

    @classmethod
    def all_known_models(cls) -> "AvailableModels":
        return cls.get_registry().available()

    @classmethod
    def available_with_local_keys(cls):
        services_with_local_keys = set(cls.key_info().select("service").to_list())
        return [
            m
            for m in cls.get_registry().available()
            if m.service_name in services_with_local_keys
        ]

    @classmethod
    def available(
        cls,
        search_term: str = None,
        name_only: bool = False,
        service: Optional[str] = None,
    ):
        # if search_term is None and service is None:
        #     print("Getting available models...")
        #     print("You have local keys for the following services:")
        #     print(cls.services_with_local_keys())
        #     print("\n")
        #     print("To see models by service, use the 'service' parameter.")
        #     print("E.g., Model.available(service='openai')")
        #     return None

        if service is not None:
            if service not in cls.services(name_only=True):
                raise ValueError(
                    f"Service {service} not found in available services.",
                    f"Available services are: {cls.services()}",
                )

        full_list = cls.get_registry().available(service=service)

        if search_term is None:
            if name_only:
                return PrettyList(
                    [m.model_name for m in full_list],
                    columns=["Model Name"],
                )
            else:
                return PrettyList(
                    [[m.model_name, m.service_name] for m in full_list],
                    columns=["Model Name", "Service Name"],
                )
        else:
            filtered_results = [
                m
                for m in full_list
                if search_term in m.model_name or search_term in m.service_name
            ]
            if name_only:
                return PrettyList(
                    [m.model_name for m in filtered_results],
                    columns=["Model Name"],
                )
            else:
                return PrettyList(
                    [[m.model_name, m.service_name] for m in full_list],
                    columns=["Model Name", "Service Name"],
                )

    @classmethod
    def check_models(cls, verbose=False):
        print("Checking all available models...\n")
        for model in cls.available(name_only=True):
            print(f"Now checking: {model}")
            try:
                m = cls(model)
            except Exception as e:
                print(f"Error creating instance of {model}: {e}")
                continue
            try:
                results = m.hello(verbose)
                if verbose:
                    print(f"Results from model call: {results}")
            except Exception as e:
                print(f"Error calling 'hello' on {model}: {e}")
                continue
            print("OK!")
            print("\n")

    @classmethod
    def example(cls, randomize: bool = False) -> "Model":
        """
        Returns an example Model instance.

        :param randomize: If True, the temperature is set to a random decimal between 0 and 1.
        """
        temperature = 0.5 if not randomize else round(random(), 2)
        model_name = cls.default_model
        return cls(model_name, temperature=temperature)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    available = Model.available()
    m = Model("gpt-4-1106-preview")
    results = m.execute_model_call("Hello world")
    print(results)



from __future__ import annotations
import asyncio
from typing import Any, Type, List, Generator, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.questions import QuestionBase
    from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
    from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
    from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog


class InterviewTaskManager:
    """Handles creation and management of interview tasks."""

    def __init__(self, survey, iteration=0):
        from edsl.jobs.tasks.TaskCreators import TaskCreators
        from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog

        self.survey = survey
        self.iteration = iteration
        self.task_creators = TaskCreators()
        self.to_index = {
            question_name: index
            for index, question_name in enumerate(self.survey.question_names)
        }
        self._task_status_log_dict = InterviewStatusLog()

    def build_question_tasks(
        self, answer_func, token_estimator, model_buckets
    ) -> list[asyncio.Task]:
        """Create tasks for all questions with proper dependencies."""
        tasks = []
        for question in self.survey.questions:
            dependencies = self._get_task_dependencies(tasks, question)
            task = self._create_single_task(
                question=question,
                dependencies=dependencies,
                answer_func=answer_func,
                token_estimator=token_estimator,
                model_buckets=model_buckets,
            )
            tasks.append(task)
        return tuple(tasks)

    def _get_task_dependencies(
        self, existing_tasks: list[asyncio.Task], question: "QuestionBase"
    ) -> list[asyncio.Task]:
        """Get tasks that must be completed before the given question."""
        dag = self.survey.dag(textify=True)
        parents = dag.get(question.question_name, [])
        return [existing_tasks[self.to_index[parent_name]] for parent_name in parents]

    def _create_single_task(
        self,
        question: "QuestionBase",
        dependencies: list[asyncio.Task],
        answer_func,
        token_estimator,
        model_buckets,
    ) -> asyncio.Task:
        """Create a single question task with its dependencies."""
        from edsl.jobs.tasks.QuestionTaskCreator import QuestionTaskCreator

        task_creator = QuestionTaskCreator(
            question=question,
            answer_question_func=answer_func,
            token_estimator=token_estimator,
            model_buckets=model_buckets,
            iteration=self.iteration,
        )

        for dependency in dependencies:
            task_creator.add_dependency(dependency)

        self.task_creators[question.question_name] = task_creator
        return task_creator.generate_task()

    @property
    def task_status_logs(self) -> "InterviewStatusLog":
        """Return the task status logs for the interview.

        The keys are the question names; the values are the lists of status log changes for each task.
        """
        for task_creator in self.task_creators.values():
            self._task_status_log_dict[
                task_creator.question.question_name
            ] = task_creator.status_log
        return self._task_status_log_dict

    @property
    def token_usage(self) -> "InterviewTokenUsage":
        """Determine how many tokens were used for the interview."""
        return self.task_creators.token_usage

    @property
    def interview_status(self) -> "InterviewStatusDictionary":
        """Return a dictionary mapping task status codes to counts."""
        return self.task_creators.interview_status



from __future__ import annotations
import time
import asyncio
import threading
import warnings
from typing import TYPE_CHECKING

from edsl.results.Results import Results
from edsl.jobs.runners.JobsRunnerStatus import JobsRunnerStatus
from edsl.jobs.tasks.TaskHistory import TaskHistory
from edsl.utilities.decorators import jupyter_nb_handler
from edsl.jobs.async_interview_runner import AsyncInterviewRunner
from edsl.jobs.data_structures import RunEnvironment, RunParameters, RunConfig

if TYPE_CHECKING:
    from edsl.jobs.Jobs import Jobs


class JobsRunnerAsyncio:
    """A class for running a collection of interviews asynchronously.

    It gets instaniated from a Jobs object.
    The Jobs object is a collection of interviews that are to be run.
    """

    def __init__(self, jobs: "Jobs", environment: RunEnvironment):
        self.jobs = jobs
        self.environment = environment

    def __len__(self):
        return len(self.jobs)

    async def run_async(self, parameters: RunParameters) -> Results:
        """Used for some other modules that have a non-standard way of running interviews."""

        self.environment.jobs_runner_status = JobsRunnerStatus(self, n=parameters.n)
        data = []
        task_history = TaskHistory(include_traceback=False)

        run_config = RunConfig(parameters=parameters, environment=self.environment)
        result_generator = AsyncInterviewRunner(self.jobs, run_config)

        async for result, interview in result_generator.run():
            data.append(result)
            task_history.add_interview(interview)

        return Results(survey=self.jobs.survey, task_history=task_history, data=data)

    def simple_run(self):
        data = asyncio.run(self.run_async())
        return Results(survey=self.jobs.survey, data=data)

    @jupyter_nb_handler
    async def run(self, parameters: RunParameters) -> Results:
        """Runs a collection of interviews, handling both async and sync contexts."""

        run_config = RunConfig(parameters=parameters, environment=self.environment)

        self.start_time = time.monotonic()
        self.completed = False

        from edsl.coop import Coop

        coop = Coop()
        endpoint_url = coop.get_progress_bar_url()

        def set_up_jobs_runner_status(jobs_runner_status):
            if jobs_runner_status is not None:
                return jobs_runner_status(
                    self,
                    n=parameters.n,
                    endpoint_url=endpoint_url,
                    job_uuid=parameters.job_uuid,
                )
            else:
                return JobsRunnerStatus(
                    self,
                    n=parameters.n,
                    endpoint_url=endpoint_url,
                    job_uuid=parameters.job_uuid,
                )

        run_config.environment.jobs_runner_status = set_up_jobs_runner_status(
            self.environment.jobs_runner_status
        )

        async def get_results(results) -> None:
            """Conducted the interviews and append to the results list."""
            result_generator = AsyncInterviewRunner(self.jobs, run_config)
            async for result, interview in result_generator.run():
                results.append(result)
                results.task_history.add_interview(interview)

            self.completed = True

        def run_progress_bar(stop_event) -> None:
            """Runs the progress bar in a separate thread."""
            self.jobs_runner_status.update_progress(stop_event)

        def set_up_progress_bar(progress_bar: bool, jobs_runner_status):
            progress_thread = None
            if progress_bar and jobs_runner_status.has_ep_api_key():
                jobs_runner_status.setup()
                progress_thread = threading.Thread(
                    target=run_progress_bar, args=(stop_event,)
                )
                progress_thread.start()
            elif progress_bar:
                warnings.warn(
                    "You need an Expected Parrot API key to view job progress bars."
                )
            return progress_thread

        results = Results(
            survey=self.jobs.survey,
            data=[],
            task_history=TaskHistory(),
            cache=self.environment.cache.new_entries_cache(),
        )
        stop_event = threading.Event()
        progress_thread = set_up_progress_bar(
            parameters.progress_bar, run_config.environment.jobs_runner_status
        )

        exception_to_raise = None
        try:
            await get_results(results)
        except KeyboardInterrupt:
            print("Keyboard interrupt received. Stopping gracefully...")
            stop_event.set()
        except Exception as e:
            if parameters.stop_on_exception:
                exception_to_raise = e
            stop_event.set()
        finally:
            stop_event.set()
            if progress_thread is not None:
                progress_thread.join()

            if exception_to_raise:
                raise exception_to_raise

            results.cache = self.environment.cache.new_entries_cache()
            results.bucket_collection = self.environment.bucket_collection

            from edsl.jobs.results_exceptions_handler import ResultsExceptionsHandler

            results_exceptions_handler = ResultsExceptionsHandler(results, parameters)

            results_exceptions_handler.handle_exceptions()
            return results



class CoopFunctionsMixin:
    def better_names(self, existing_names):
        from edsl import QuestionList, Scenario

        s = Scenario({"existing_names": existing_names})
        q = QuestionList(
            question_text="""The following colum names are already in use: {{ existing_names }} 
                         Please provide new names for the columns.
                         They should be short, one or two words, and unique. They should be valid Python idenifiers. 
                         No spaces - use underscores instead. 
                         """,
            question_name="better_names",
        )
        results = q.by(s).run(verbose=False)
        return results.select("answer.better_names").first()



from edsl.exceptions.BaseException import BaseException


class SurveyError(BaseException):
    relevant_doc = "https://docs.expectedparrot.com/en/latest/surveys.html"


class SurveyCreationError(SurveyError):
    pass


class SurveyHasNoRulesError(SurveyError):
    pass


class SurveyRuleSendsYouBackwardsError(SurveyError):
    pass


class SurveyRuleSkipLogicSyntaxError(SurveyError):
    pass


class SurveyRuleReferenceInRuleToUnknownQuestionError(SurveyError):
    pass


class SurveyRuleRefersToFutureStateError(SurveyError):
    pass


class SurveyRuleCollectionHasNoRulesAtNodeError(SurveyError):
    pass


class SurveyRuleCannotEvaluateError(SurveyError):
    pass



from typing import Optional, Union, Literal, TYPE_CHECKING, NewType, Callable, Any

from dataclasses import dataclass


Seconds = NewType("Seconds", float)
JobUUID = NewType("JobUUID", str)

from edsl.exceptions.coop import CoopServerResponseError

if TYPE_CHECKING:
    from edsl.results.Results import Results
    from edsl.jobs.Jobs import Jobs
    from edsl.coop.coop import RemoteInferenceResponse, RemoteInferenceCreationInfo
    from edsl.jobs.JobsRemoteInferenceLogger import JobLogger

from edsl.coop.coop import RemoteInferenceResponse, RemoteInferenceCreationInfo

from edsl.jobs.jobs_status_enums import JobsStatus
from edsl.coop.utils import VisibilityType
from edsl.jobs.JobsRemoteInferenceLogger import JobLogger


class RemoteJobConstants:
    """Constants for remote job handling."""

    REMOTE_JOB_POLL_INTERVAL = 1
    REMOTE_JOB_VERBOSE = False
    DISCORD_URL = "https://discord.com/invite/mxAYkjfy9m"


@dataclass
class RemoteJobInfo:
    creation_data: RemoteInferenceCreationInfo
    job_uuid: JobUUID
    logger: JobLogger


class JobsRemoteInferenceHandler:
    def __init__(
        self,
        jobs: "Jobs",
        verbose: bool = RemoteJobConstants.REMOTE_JOB_VERBOSE,
        poll_interval: Seconds = RemoteJobConstants.REMOTE_JOB_POLL_INTERVAL,
    ):
        """Handles the creation and running of a remote inference job."""
        self.jobs = jobs
        self.verbose = verbose
        self.poll_interval = poll_interval

        from edsl.config import CONFIG

        self.expected_parrot_url = CONFIG.get("EXPECTED_PARROT_URL")
        self.remote_inference_url = f"{self.expected_parrot_url}/home/remote-inference"

    def _create_logger(self) -> JobLogger:
        from edsl.utilities.is_notebook import is_notebook
        from edsl.jobs.JobsRemoteInferenceLogger import (
            JupyterJobLogger,
            StdOutJobLogger,
        )
        from edsl.jobs.loggers.HTMLTableJobLogger import HTMLTableJobLogger

        if is_notebook():
            return HTMLTableJobLogger(verbose=self.verbose)
        return StdOutJobLogger(verbose=self.verbose)

    def use_remote_inference(self, disable_remote_inference: bool) -> bool:
        import requests

        if disable_remote_inference:
            return False
        if not disable_remote_inference:
            try:
                from edsl.coop.coop import Coop

                user_edsl_settings = Coop().edsl_settings
                return user_edsl_settings.get("remote_inference", False)
            except requests.ConnectionError:
                pass
            except CoopServerResponseError as e:
                pass

        return False

    def create_remote_inference_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional[VisibilityType] = "unlisted",
    ) -> RemoteJobInfo:

        from edsl.config import CONFIG
        from edsl.coop.coop import Coop

        logger = self._create_logger()

        coop = Coop()
        logger.update(
            "Remote inference activated. Sending job to server...",
            status=JobsStatus.QUEUED,
        )
        remote_job_creation_data = coop.remote_inference_create(
            self.jobs,
            description=remote_inference_description,
            status="queued",
            iterations=iterations,
            initial_results_visibility=remote_inference_results_visibility,
        )
        logger.update(
            "Your survey is running at the Expected Parrot server...",
            status=JobsStatus.RUNNING,
        )
        job_uuid = remote_job_creation_data.get("uuid")
        logger.update(
            message=f"Job sent to server. (Job uuid={job_uuid}).",
            status=JobsStatus.RUNNING,
        )
        logger.add_info("job_uuid", job_uuid)

        logger.update(
            f"Job details are available at your Coop account {self.remote_inference_url}",
            status=JobsStatus.RUNNING,
        )
        progress_bar_url = (
            f"{self.expected_parrot_url}/home/remote-job-progress/{job_uuid}"
        )
        logger.add_info("progress_bar_url", progress_bar_url)
        logger.update(
            f"View job progress here: {progress_bar_url}", status=JobsStatus.RUNNING
        )

        return RemoteJobInfo(
            creation_data=remote_job_creation_data,
            job_uuid=job_uuid,
            logger=logger,
        )

    @staticmethod
    def check_status(
        job_uuid: JobUUID,
    ) -> RemoteInferenceResponse:
        from edsl.coop.coop import Coop

        coop = Coop()
        return coop.remote_inference_get(job_uuid)

    def _construct_remote_job_fetcher(
        self, testing_simulated_response: Optional[Any] = None
    ) -> Callable:
        if testing_simulated_response is not None:
            return lambda job_uuid: testing_simulated_response
        else:
            from edsl.coop.coop import Coop

            coop = Coop()
            return coop.remote_inference_get

    def _construct_object_fetcher(
        self, testing_simulated_response: Optional[Any] = None
    ) -> Callable:
        "Constructs a function to fetch the results object from Coop."
        if testing_simulated_response is not None:
            return lambda results_uuid, expected_object_type: Results.example()
        else:
            from edsl.coop.coop import Coop

            coop = Coop()
            return coop.get

    def _handle_cancelled_job(self, job_info: RemoteJobInfo) -> None:
        "Handles a cancelled job by logging the cancellation and updating the job status."

        job_info.logger.update(
            message="Job cancelled by the user.", status=JobsStatus.CANCELLED
        )
        job_info.logger.update(
            f"See {self.expected_parrot_url}/home/remote-inference for more details.",
            status=JobsStatus.CANCELLED,
        )

    def _handle_failed_job(
        self, job_info: RemoteJobInfo, remote_job_data: RemoteInferenceResponse
    ) -> None:
        "Handles a failed job by logging the error and updating the job status."
        latest_error_report_url = remote_job_data.get("latest_error_report_url")
        if latest_error_report_url:
            job_info.logger.add_info("error_report_url", latest_error_report_url)

        job_info.logger.update("Job failed.", status=JobsStatus.FAILED)
        job_info.logger.update(
            f"See {self.expected_parrot_url}/home/remote-inference for more details.",
            status=JobsStatus.FAILED,
        )
        job_info.logger.update(
            f"Need support? Visit Discord: {RemoteJobConstants.DISCORD_URL}",
            status=JobsStatus.FAILED,
        )

    def _sleep_for_a_bit(self, job_info: RemoteJobInfo, status: str) -> None:
        import time
        from datetime import datetime

        time_checked = datetime.now().strftime("%Y-%m-%d %I:%M:%S %p")
        job_info.logger.update(
            f"Job status: {status} - last update: {time_checked}",
            status=JobsStatus.RUNNING,
        )
        time.sleep(self.poll_interval)

    def _fetch_results_and_log(
        self,
        job_info: RemoteJobInfo,
        results_uuid: str,
        remote_job_data: RemoteInferenceResponse,
        object_fetcher: Callable,
    ) -> "Results":
        "Fetches the results object and logs the results URL."
        job_info.logger.add_info("results_uuid", results_uuid)
        results = object_fetcher(results_uuid, expected_object_type="results")
        results_url = remote_job_data.get("results_url")
        job_info.logger.update(
            f"Job completed and Results stored on Coop: {results_url}",
            status=JobsStatus.COMPLETED,
        )
        results.job_uuid = job_info.job_uuid
        results.results_uuid = results_uuid
        return results

    def poll_remote_inference_job(
        self,
        job_info: RemoteJobInfo,
        testing_simulated_response=None,
    ) -> Union[None, "Results"]:
        """Polls a remote inference job for completion and returns the results."""

        remote_job_data_fetcher = self._construct_remote_job_fetcher(
            testing_simulated_response
        )
        object_fetcher = self._construct_object_fetcher(testing_simulated_response)

        job_in_queue = True
        while job_in_queue:
            remote_job_data = remote_job_data_fetcher(job_info.job_uuid)
            status = remote_job_data.get("status")

            if status == "cancelled":
                self._handle_cancelled_job(job_info)
                return None

            elif status == "failed" or status == "completed":
                if status == "failed":
                    self._handle_failed_job(job_info, remote_job_data)

                results_uuid = remote_job_data.get("results_uuid")
                if results_uuid:
                    results = self._fetch_results_and_log(
                        job_info=job_info,
                        results_uuid=results_uuid,
                        remote_job_data=remote_job_data,
                        object_fetcher=object_fetcher,
                    )
                    return results
                else:
                    return None

            else:
                self._sleep_for_a_bit(job_info, status)

    async def create_and_poll_remote_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional[VisibilityType] = "unlisted",
    ) -> Union["Results", None]:
        """
        Creates and polls a remote inference job asynchronously.
        Reuses existing synchronous methods but runs them in an async context.

        :param iterations: Number of times to run each interview
        :param remote_inference_description: Optional description for the remote job
        :param remote_inference_results_visibility: Visibility setting for results
        :return: Results object if successful, None if job fails or is cancelled
        """
        import asyncio
        from functools import partial

        # Create job using existing method
        loop = asyncio.get_event_loop()
        job_info = await loop.run_in_executor(
            None,
            partial(
                self.create_remote_inference_job,
                iterations=iterations,
                remote_inference_description=remote_inference_description,
                remote_inference_results_visibility=remote_inference_results_visibility,
            ),
        )
        if job_info is None:
            raise ValueError("Remote job creation failed.")

        return await loop.run_in_executor(
            None,
            partial(self.poll_remote_inference_job, job_info),
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
from typing import Optional
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

from edsl.questions.decorators import inject_exception


class QuestionLikertFive(QuestionMultipleChoice):
    """This question prompts the agent to respond to a statement on a 5-point Likert scale."""

    question_type = "likert_five"
    likert_options: list[str] = [
        "Strongly disagree",
        "Disagree",
        "Neutral",
        "Agree",
        "Strongly agree",
    ]

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: Optional[list[str]] = likert_options,
        answering_instructions: Optional[str] = None,
        question_presentation: Optional[str] = None,
        include_comment: bool = True,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options the respondent should select from (list of strings). If not provided, the default Likert options are used (['Strongly disagree', 'Disagree', 'Neutral', 'Agree', 'Strongly agree']). To view them, run `QuestionLikertFive.likert_options`.
        """
        super().__init__(
            question_name=question_name,
            question_text=question_text,
            question_options=question_options,
            use_code=False,
            include_comment=include_comment,
            answering_instructions=answering_instructions,
            question_presentation=question_presentation,
        )

    @classmethod
    @inject_exception
    def example(cls) -> QuestionLikertFive:
        """Return an example question."""
        return cls(
            question_name="happy_raining",
            question_text="I'm only happy when it rains.",
        )


def main():
    """Test QuestionLikertFive."""
    from edsl.questions.derived.QuestionLikertFive import QuestionLikertFive

    q = QuestionLikertFive.example()
    q.question_text
    q.question_options
    q.question_name
    # validate an answer
    q._validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer(0, {})
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from abc import ABC, abstractmethod
import asyncio
from typing import Coroutine, Dict, Any, Optional, TYPE_CHECKING

from edsl.utilities.decorators import jupyter_nb_handler
from edsl.data_transfer_models import AgentResponseDict

if TYPE_CHECKING:
    from edsl.prompts.Prompt import Prompt
    from edsl.data.Cache import Cache
    from edsl.questions.QuestionBase import QuestionBase
    from edsl.scenarios.Scenario import Scenario
    from edsl.surveys.MemoryPlan import MemoryPlan
    from edsl.language_models.LanguageModel import LanguageModel
    from edsl.surveys.Survey import Survey
    from edsl.agents.Agent import Agent
    from edsl.language_models.key_management.KeyLookup import KeyLookup

from edsl.data_transfer_models import EDSLResultObjectInput
from edsl.agents.PromptConstructor import PromptConstructor
from edsl.agents.prompt_helpers import PromptPlan


class InvigilatorBase(ABC):
    """An invigiator (someone who administers an exam) is a class that is responsible for administering a question to an agent.

    >>> InvigilatorBase.example().answer_question()
    {'message': [{'text': 'SPAM!'}], 'usage': {'prompt_tokens': 1, 'completion_tokens': 1}}

    >>> InvigilatorBase.example().get_failed_task_result(failure_reason="Failed to get response").comment
    'Failed to get response'

    This returns an empty prompt because there is no memory the agent needs to have at q0.
    """

    def __init__(
        self,
        agent: "Agent",
        question: "QuestionBase",
        scenario: "Scenario",
        model: "LanguageModel",
        memory_plan: "MemoryPlan",
        current_answers: dict,
        survey: Optional["Survey"],
        cache: Optional["Cache"] = None,
        iteration: Optional[int] = 1,
        additional_prompt_data: Optional[dict] = None,
        raise_validation_errors: Optional[bool] = True,
        prompt_plan: Optional["PromptPlan"] = None,
        key_lookup: Optional["KeyLookup"] = None,
    ):
        """Initialize a new Invigilator."""
        self.agent = agent
        self.question = question
        self.scenario = scenario
        self.model = model
        self.memory_plan = memory_plan
        self.current_answers = current_answers or {}
        self.iteration = iteration
        self.additional_prompt_data = additional_prompt_data
        self.cache = cache
        self.survey = survey
        self.raise_validation_errors = raise_validation_errors
        self.key_lookup = key_lookup

        if prompt_plan is None:
            self.prompt_plan = PromptPlan()
        else:
            self.prompt_plan = prompt_plan

        # placeholder to store the raw model response
        self.raw_model_response = None

    @property
    def prompt_constructor(self) -> PromptConstructor:
        """Return the prompt constructor."""
        return PromptConstructor(self, prompt_plan=self.prompt_plan)

    def to_dict(self, include_cache=False) -> Dict[str, Any]:
        attributes = [
            "agent",
            "question",
            "scenario",
            "model",
            "memory_plan",
            "current_answers",
            "iteration",
            "additional_prompt_data",
            "survey",
        ]
        if include_cache:
            attributes.append("cache")

        def serialize_attribute(attr):
            value = getattr(self, attr)
            if value is None:
                return None
            if hasattr(value, "to_dict"):
                return value.to_dict()
            if isinstance(value, (int, float, str, bool, dict, list)):
                return value
            return str(value)

        return {attr: serialize_attribute(attr) for attr in attributes}

    @classmethod
    def from_dict(cls, data) -> "InvigilatorBase":
        from edsl.agents.Agent import Agent
        from edsl.questions import QuestionBase
        from edsl.scenarios.Scenario import Scenario
        from edsl.surveys.MemoryPlan import MemoryPlan
        from edsl.language_models.LanguageModel import LanguageModel
        from edsl.surveys.Survey import Survey
        from edsl.data.Cache import Cache

        attributes_to_classes = {
            "agent": Agent,
            "question": QuestionBase,
            "scenario": Scenario,
            "model": LanguageModel,
            "memory_plan": MemoryPlan,
            "survey": Survey,
            "cache": Cache,
        }
        d = {}
        for attr, cls_ in attributes_to_classes.items():
            if attr in data and data[attr] is not None:
                if attr not in data:
                    d[attr] = {}
                else:
                    d[attr] = cls_.from_dict(data[attr])

        d["current_answers"] = data["current_answers"]
        d["iteration"] = data["iteration"]
        d["additional_prompt_data"] = data["additional_prompt_data"]

        d = cls(**d)

    def __repr__(self) -> str:
        """Return a string representation of the Invigilator.

        >>> InvigilatorBase.example().__repr__()
        'InvigilatorExample(...)'

        """
        return f"{self.__class__.__name__}(agent={repr(self.agent)}, question={repr(self.question)}, scneario={repr(self.scenario)}, model={repr(self.model)}, memory_plan={repr(self.memory_plan)}, current_answers={repr(self.current_answers)}, iteration{repr(self.iteration)}, additional_prompt_data={repr(self.additional_prompt_data)}, cache={repr(self.cache)})"

    def get_failed_task_result(self, failure_reason: str) -> EDSLResultObjectInput:
        """Return an AgentResponseDict used in case the question-asking fails.

        Possible reasons include:
        - Legimately skipped because of skip logic
        - Failed to get response from the model

        """
        data = {
            "answer": None,
            "generated_tokens": None,
            "comment": failure_reason,
            "question_name": self.question.question_name,
            "prompts": self.get_prompts(),
            "cached_response": None,
            "raw_model_response": None,
            "cache_used": None,
            "cache_key": None,
        }
        return EDSLResultObjectInput(**data)

    def get_prompts(self) -> Dict[str, "Prompt"]:
        """Return the prompt used."""
        from edsl.prompts.Prompt import Prompt

        return {
            "user_prompt": Prompt("NA"),
            "system_prompt": Prompt("NA"),
        }

    @abstractmethod
    async def async_answer_question(self):
        """Asnwer a question."""
        pass

    @jupyter_nb_handler
    def answer_question(self) -> Coroutine:
        """Return a function that gets the answers to the question."""

        async def main():
            """Return the answer to the question."""
            results = await asyncio.gather(self.async_answer_question())
            return results[0]  # Since there's only one task, return its result

        return main()

    @classmethod
    def example(
        cls, throw_an_exception=False, question=None, scenario=None, survey=None
    ) -> "InvigilatorBase":
        """Return an example invigilator.

        >>> InvigilatorBase.example()
        InvigilatorExample(...)

        >>> InvigilatorBase.example().answer_question()
        {'message': [{'text': 'SPAM!'}], 'usage': {'prompt_tokens': 1, 'completion_tokens': 1}}

        >>> InvigilatorBase.example(throw_an_exception=True).answer_question()
        Traceback (most recent call last):
        ...
        Exception: This is a test error
        """
        from edsl.agents.Agent import Agent
        from edsl.scenarios.Scenario import Scenario
        from edsl.surveys.MemoryPlan import MemoryPlan
        from edsl.language_models.model import Model
        from edsl.surveys.Survey import Survey

        model = Model("test", canned_response="SPAM!")

        if throw_an_exception:
            model.throw_exception = True
        agent = Agent.example()

        if not survey:
            survey = Survey.example()

        if question not in survey.questions and question is not None:
            survey.add_question(question)

        question = question or survey.questions[0]
        scenario = scenario or Scenario.example()
        memory_plan = MemoryPlan(survey=survey)
        current_answers = None

        class InvigilatorExample(cls):
            """An example invigilator."""

            async def async_answer_question(self):
                """Answer a question."""
                return await self.model.async_execute_model_call(
                    user_prompt="Hello", system_prompt="Hi"
                )

        return InvigilatorExample(
            agent=agent,
            question=question,
            scenario=scenario,
            survey=survey,
            model=model,
            memory_plan=memory_plan,
            current_answers=current_answers,
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
from typing import Union, Optional, Dict, List, Any

from pydantic import BaseModel, Field, field_validator
from jinja2 import Template
import random
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import (
    QuestionOptionsDescriptor,
    OptionLabelDescriptor,
    QuestionTextDescriptor,
)
from edsl.questions.response_validator_abc import ResponseValidatorABC
from edsl.questions.decorators import inject_exception
from edsl.exceptions.questions import (
    QuestionAnswerValidationError,
    QuestionCreationValidationError,
)


def create_matrix_response(
    question_items: List[str],
    question_options: List[Union[int, str, float]],
    permissive: bool = False,
):
    """Create a response model for matrix questions.

    The response model validates that:
    1. All question items are answered
    2. Each answer is from the allowed options
    """

    if permissive:

        class MatrixResponse(BaseModel):
            answer: Dict[str, Any]
            comment: Optional[str] = None
            generated_tokens: Optional[Any] = None

    else:

        class MatrixResponse(BaseModel):
            answer: Dict[str, Union[int, str, float]] = Field(
                ..., description="Mapping of items to selected options"
            )
            comment: Optional[str] = None
            generated_tokens: Optional[Any] = None

            @field_validator("answer")
            def validate_answer(cls, v, values, **kwargs):
                # Check that all items have responses
                if not all(item in v for item in question_items):
                    missing = set(question_items) - set(v.keys())
                    raise ValueError(f"Missing responses for items: {missing}")

                # Check that all responses are valid options
                if not all(answer in question_options for answer in v.values()):
                    invalid = [ans for ans in v.values() if ans not in question_options]
                    raise ValueError(f"Invalid options selected: {invalid}")
                return v

    return MatrixResponse


class MatrixResponseValidator(ResponseValidatorABC):
    required_params = ["question_items", "question_options", "permissive"]

    valid_examples = [
        (
            {"answer": {"Item1": 1, "Item2": 2}},
            {
                "question_items": ["Item1", "Item2"],
                "question_options": [1, 2, 3],
            },
        )
    ]

    invalid_examples = [
        (
            {"answer": {"Item1": 1}},
            {
                "question_items": ["Item1", "Item2"],
                "question_options": [1, 2, 3],
            },
            "Missing responses for some items",
        ),
        (
            {"answer": {"Item1": 4, "Item2": 5}},
            {
                "question_items": ["Item1", "Item2"],
                "question_options": [1, 2, 3],
            },
            "Invalid options selected",
        ),
    ]

    def fix(self, response, verbose=False):
        if verbose:
            print(f"Fixing matrix response: {response}")

        # If we have generated tokens, try to parse them
        if "generated_tokens" in response:
            try:
                import json

                fixed = json.loads(response["generated_tokens"])
                if isinstance(fixed, dict):
                    # Map numeric keys to question items
                    mapped_answer = {}
                    for idx, item in enumerate(self.question_items):
                        if str(idx) in fixed:
                            mapped_answer[item] = fixed[str(idx)]
                    if (
                        mapped_answer
                    ):  # Only return if we successfully mapped some answers
                        return {"answer": mapped_answer}
            except:
                pass

        # If answer uses numeric keys, map them to question items
        if "answer" in response and isinstance(response["answer"], dict):
            if all(str(key).isdigit() for key in response["answer"].keys()):
                mapped_answer = {}
                for idx, item in enumerate(self.question_items):
                    if str(idx) in response["answer"]:
                        mapped_answer[item] = response["answer"][str(idx)]
                if mapped_answer:  # Only update if we successfully mapped some answers
                    response["answer"] = mapped_answer

        return response


class QuestionMatrix(QuestionBase):
    """A question that presents a matrix/grid where multiple items are rated using the same scale."""

    question_type = "matrix"
    question_text: str = QuestionTextDescriptor()
    question_items: List[str] = QuestionOptionsDescriptor()
    question_options: List[Union[int, str, float]] = QuestionOptionsDescriptor()
    option_labels: Optional[Dict[Union[int, str, float], str]] = OptionLabelDescriptor()

    _response_model = None
    response_validator_class = MatrixResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_items: List[str],
        question_options: List[Union[int, str, float]],
        option_labels: Optional[Dict[Union[int, str, float], str]] = None,
        include_comment: bool = True,
        answering_instructions: Optional[str] = None,
        question_presentation: Optional[str] = None,
        permissive: bool = False,
    ):
        """Initialize a matrix question.

        Args:
            question_name: The name of the question
            question_text: The text of the question
            question_items: List of items to be rated
            question_options: List of rating options
            option_labels: Optional mapping of options to their labels
            include_comment: Whether to include a comment field
            answering_instructions: Optional custom instructions
            question_presentation: Optional custom presentation
            permissive: Whether to strictly validate responses
        """
        self.question_name = question_name

        try:
            self.question_text = question_text
        except Exception as e:
            raise QuestionCreationValidationError(
                "question_text cannot be empty or too short!"
            ) from e

        self.question_items = question_items
        self.question_options = question_options
        self.option_labels = option_labels or {}

        self.include_comment = include_comment
        self.answering_instructions = answering_instructions
        self.question_presentation = question_presentation
        self.permissive = permissive

    def create_response_model(self):
        return create_matrix_response(
            self.question_items, self.question_options, self.permissive
        )

    @property
    def question_html_content(self) -> str:
        """Generate HTML representation of the matrix question."""
        template = Template(
            """
        <table class="matrix-question">
            <tr>
                <th></th>
                {% for option in question_options %}
                <th>
                    {{ option }}
                    {% if option in option_labels %}
                    <br>
                    <small>{{ option_labels[option] }}</small>
                    {% endif %}
                </th>
                {% endfor %}
            </tr>
            {% for item in question_items %}
            <tr>
                <td>{{ item }}</td>
                {% for option in question_options %}
                <td>
                    <input type="radio" 
                           name="{{ question_name }}_{{ item }}" 
                           value="{{ option }}"
                           id="{{ question_name }}_{{ item }}_{{ option }}">
                </td>
                {% endfor %}
            </tr>
            {% endfor %}
        </table>
        """
        )

        return template.render(
            question_name=self.question_name,
            question_items=self.question_items,
            question_options=self.question_options,
            option_labels=self.option_labels,
        )

    @classmethod
    @inject_exception
    def example(cls) -> QuestionMatrix:
        """Return an example matrix question."""
        return cls(
            question_name="child_happiness",
            question_text="How happy would you be with different numbers of children?",
            question_items=[
                "No children",
                "1 child",
                "2 children",
                "3 or more children",
            ],
            question_options=[1, 2, 3, 4, 5],
            option_labels={1: "Very sad", 3: "Neutral", 5: "Extremely happy"},
        )

    def _simulate_answer(self) -> dict:
        """Simulate a random valid answer."""
        return {
            "answer": {
                item: random.choice(self.question_options)
                for item in self.question_items
            }
        }


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
import json
import re

from typing import Any, Optional, Dict
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import AnswerTemplateDescriptor

from edsl.questions.response_validator_abc import ResponseValidatorABC
from edsl.questions.data_structures import BaseResponse
from edsl.questions.decorators import inject_exception

from typing import Dict, Any
from pydantic import create_model, Field


def extract_json(text, expected_keys, verbose=False):
    # Escape special regex characters in keys
    escaped_keys = [re.escape(key) for key in expected_keys]

    # Create a pattern that looks for all expected keys
    pattern = r"\{[^}]*" + r"[^}]*".join(escaped_keys) + r"[^}]*\}"

    json_match = re.search(pattern, text)

    if json_match:
        json_str = json_match.group(0)
        try:
            # Parse the extracted string as JSON
            json_data = json.loads(json_str)

            # Verify that all expected keys are present
            if all(key in json_data for key in expected_keys):
                return json_data
            else:
                if verbose:
                    print(
                        "Error: Not all expected keys were found in the extracted JSON."
                    )
                return None
        except json.JSONDecodeError:
            if verbose:
                print("Error: The extracted content is not valid JSON.")
            return None
    else:
        if verbose:
            print("Error: No JSON-like structure found with all expected keys.")
        return None


def dict_to_pydantic_model(input_dict: Dict[str, Any]) -> Any:
    field_definitions = {
        key: (str, Field(default=str(value))) for key, value in input_dict.items()
    }

    DynamicModel = create_model("DynamicModel", **field_definitions)

    class AnswerModel(BaseResponse):
        answer: "DynamicModel"
        generated_tokens: Optional[str] = None
        comment: Optional[str] = None

    return AnswerModel


class ExtractResponseValidator(ResponseValidatorABC):
    required_params = ["answer_template"]
    valid_examples = [({"answer": "This is great"}, {})]
    invalid_examples = [
        (
            {"answer": None},
            {"answer_template": {"name": "John Doe", "profession": "Carpenter"}},
            "Result cannot be empty",
        ),
    ]

    def custom_validate(self, response) -> BaseResponse:
        return response.dict()

    def fix(self, response, verbose=False):
        raw_tokens = response["generated_tokens"]
        if verbose:
            print(f"Invalid response of QuestionExtract was: {raw_tokens}")
        extracted_json = extract_json(raw_tokens, self.answer_template.keys(), verbose)
        if verbose:
            print("Proposed solution is: ", extracted_json)
        return {
            "answer": extracted_json,
            "comment": response.get("comment", None),
            "generated_tokens": raw_tokens,
        }


class QuestionExtract(QuestionBase):
    """This question prompts the agent to extract information from a string and return it in a given template."""

    question_type = "extract"
    answer_template: dict[str, Any] = AnswerTemplateDescriptor()
    _response_model = None
    response_validator_class = ExtractResponseValidator

    def __init__(
        self,
        question_text: str,
        answer_template: dict[str, Any],
        question_name: str,
        answering_instructions: str = None,
        question_presentation: str = None,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param answer_template: The template for the answer.
        :param answering_instructions: Instructions for answering the question.
        :param question_presentation: The presentation of the question.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.answer_template = answer_template
        self.answering_instructions = answering_instructions
        self.question_presentation = question_presentation

    def create_response_model(self):
        return dict_to_pydantic_model(self.answer_template)

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        {% for field, placeholder in answer_template.items() %}
        <div>
        <label for="{{ field }}">{{ field }}</label>
        <input type="text" id="{{ field }}" name="{{ question_name }}[{{ field }}]" placeholder="{{ placeholder }}">
        </div>
        {% endfor %}
        """
        ).render(
            question_name=self.question_name,
            answer_template=self.answer_template,
        )
        return question_html_content

    @classmethod
    @inject_exception
    def example(cls) -> QuestionExtract:
        """Return an example question."""
        return cls(
            question_name="extract_name",
            question_text="My name is Moby Dick. I have a PhD in astrology, but I'm actually a truck driver",
            answer_template={"name": "John Doe", "profession": "Carpenter"},
        )


def main():
    """Administer a question and validate the answer."""
    from edsl.questions.QuestionExtract import QuestionExtract

    q = QuestionExtract.example()
    q.question_text
    q.question_name
    q.answer_template
    q._validate_answer({"answer": {"name": "Moby", "profession": "truck driver"}})
    q._translate_answer_code_to_answer(
        {"answer": {"name": "Moby", "profession": "truck driver"}}
    )
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
